/*

    File: dir.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#include <stdio.h>
#include <ctype.h>	/* toupper */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "intrface.h"
#include "dir.h"
static const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

long int dir_aff(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, t_file_data*dir_list, const unsigned long int cluster)
{
  /* Return value
   * -1: quit
   *  1: back
   *  other: new cluster
   * */
  int quit;
  int offset=0;
  int pos_num=0;
  t_file_data *current_file;
  t_file_data *pos=dir_list;
  /* Log file */
  aff_part_rapport(1,disk_car,partition);
  ecrit_rapport("\n");
  for(current_file=dir_list;current_file;current_file=current_file->next)
  {
    dir_data->write_rapport(current_file->file_data);
  }
  /* */
  aff_copy(dir_data->window);
  wmove(dir_data->window,4,0);
  aff_part(dir_data->window,0,disk_car,partition);
  mvwaddstr(dir_data->window,5,0,"Use arrow keys to change directory, enter to quit");
  do
  {
    int i;
    int car;
    for(i=0,current_file=dir_list;(current_file!=NULL) && (i<offset);current_file=current_file->next,i++);
    for(i=offset;(current_file!=NULL) &&((i-offset)<INTER_STRUCTURE);i++,current_file=current_file->next)
    {
      wmove(dir_data->window,8+i-offset,0);
      wclrtoeol(dir_data->window);	/* before addstr for BSD compatibility */
      if(current_file==pos)
      {
	wstandout(dir_data->window);
	dir_data->write_display(dir_data->window,current_file->file_data);
	wstandend(dir_data->window);
      } else
	dir_data->write_display(dir_data->window,current_file->file_data);
    }
    wrefresh(dir_data->window);
    car=wgetch(dir_data->window);
    quit=0;
    switch(car)
    {
      case 'q':
      case '\r':
      case '\n':
      case 'M':
      case KEY_ENTER:
#ifdef DJGPP
      case PADENTER:
#endif
	quit=1;
	break;
    }
    if(dir_list)
    {
      switch(car)
      {
	case KEY_UP:
	  if(pos->prev)
	  {
	    pos=pos->prev;
	    pos_num--;
	  }
	  if(pos_num<offset)
	    offset--;
	  break;
	case KEY_DOWN:
	  if(pos->next)
	  {
	    pos=pos->next;
	    pos_num++;
	  }
	  if(pos_num>=offset+INTER_STRUCTURE)
	    offset++;
	  break;
	case '+':
	case ' ':
	case KEY_RIGHT:
	  if((pos!=NULL) && (dir_data->is_directory(pos->file_data)!=0))
	  {
	    unsigned long int new_cluster=dir_data->get_cluster_inode(pos->file_data);
	    if(new_cluster!=cluster)
	      return (long int)new_cluster;
	  }
	  break;
	case '-':
	case KEY_LEFT:
	  return 1;
	case KEY_PPAGE:
	  for(i=0;(i<INTER_STRUCTURE-1)&&(pos->prev);i++)
	  {
	    pos=pos->prev;
	    pos_num--;
	    if(pos_num<offset)
	      offset--;
	  }
	  break;
	case KEY_NPAGE:
	  for(i=0;(i<INTER_STRUCTURE-1)&&(pos->next);i++)
	  {
	    pos=pos->next;
	    pos_num++;
	    if(pos_num>=offset+INTER_STRUCTURE)
	      offset++;
	  }
	  break;
      }
    }
  } while(quit==0);
  return -1;
}

void delete_list_file(t_file_data *file_list)
{
  t_file_data *current_file=file_list;
  /* Libere la memoire */
  while(current_file)
  {
    t_file_data *next=current_file->next;
    FREE(current_file->file_data);
    FREE(current_file);
    current_file=next;
  }
}

int dir_partition(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int cluster)
{
  t_file_data *dir_list;
  long int new_cluster;
//  if(cluster<0)
//    return cluster;
  if(dir_data->debug!=0)
    ecrit_rapport("\ndir_partition cluster=%ld\n",cluster);
  dir_list=dir_data->get_dir(disk_car,partition,dir_data,cluster);
  do
  {
    new_cluster=dir_aff(disk_car,partition,dir_data,dir_list,cluster);
    if(new_cluster==0 || new_cluster>1)
    {
      if(dir_partition(disk_car, partition, dir_data, (unsigned long int)new_cluster)<0)
      { /* quit */
	delete_list_file(dir_list);
	return -1;
      }
      /* back */
    }
  } while(new_cluster==0 || new_cluster>1);
  delete_list_file(dir_list);
  return new_cluster;
}

