/*

    File: fat.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdio.h>
#include <ctype.h>	/* toupper */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "intrface.h"


int dump_fat_rapport(const struct fat_boot_sector*fh1, const int upart_type)
{
  ecrit_rapport("sector_size  %u\n", sector_size(fh1));
  ecrit_rapport("cluster_size %u\n", fh1->cluster_size);
  ecrit_rapport("reserved     %u\n", fh1->reserved);
  ecrit_rapport("fats         %u\n", fh1->fats);
  ecrit_rapport("dir_entries  %u\n", get_dir_entries(fh1));
  ecrit_rapport("sectors      %u\n", sectors(fh1));
  ecrit_rapport("media        %02X\n", fh1->media);
  ecrit_rapport("fat_length   %u\n", fh1->fat_length);
  ecrit_rapport("secs_track   %u\n", fh1->secs_track);
  ecrit_rapport("heads        %u\n", fh1->heads);
  ecrit_rapport("hidden       %u\n", fh1->hidden);
  ecrit_rapport("total_sect   %u\n", fh1->total_sect);
  if(upart_type==UP_FAT32)
  {
      ecrit_rapport("fat32_length %u\n", fh1->fat32_length);
      ecrit_rapport("flags        %04X\n", fh1->flags);
      ecrit_rapport("version      %u.%u\n", fh1->version[0], fh1->version[1]);
      ecrit_rapport("root_cluster %u\n", fh1->root_cluster);
      ecrit_rapport("info_sector  %u\n", fh1->info_sector);
      ecrit_rapport("backup_boot  %u\n", fh1->backup_boot);
  }
  return 0;
}

int dump_fat_info(const struct fat_boot_sector*fh1, const int upart_type)
{
  switch(upart_type)
  {
    case UP_FAT12:
      doprintf("FAT : 12\n");
      break;
    case UP_FAT16:
      doprintf("FAT : 16\n");
      break;
    case UP_FAT32:
      doprintf("FAT : 32\n");
      break;
    default:
      doprintf("Not a FAT\n");
      return 0;
  }
  doprintf("cluster_size %u\n", fh1->cluster_size);
  if(sectors(fh1)!=0)
    doprintf("sectors      %u\n", sectors(fh1));
  if(fh1->total_sect!=0)
    doprintf("total_sect   %u\n", fh1->total_sect);
  if(upart_type==UP_FAT32)
  {
    doprintf("fat32_length %u\n", fh1->fat32_length);
    doprintf("root_cluster %u\n", fh1->root_cluster);
    doprintf("flags        %04X\n", fh1->flags);
    doprintf("version      %u.%u\n", fh1->version[0], fh1->version[1]);
    doprintf("root_cluster %u\n", fh1->root_cluster);
    doprintf("info_sector  %u\n", fh1->info_sector);
    doprintf("backup_boot  %u\n", fh1->backup_boot);
  } else {
    doprintf("fat_length   %u\n", fh1->fat_length);
    doprintf("dir_entries  %u\n", get_dir_entries(fh1));
  }
  return 0;
}

int dump_2fat_info(const struct fat_boot_sector*fh1, const struct fat_boot_sector*fh2, const int upart_type)
{
  switch(upart_type)
  {
    case UP_FAT12:
      doprintf("FAT : 12\n");
      break;
    case UP_FAT16:
      doprintf("FAT : 16\n");
      break;
    case UP_FAT32:
      doprintf("FAT : 32\n");
      break;
    default:
      doprintf("Not a FAT\n");
      return 0;
  }
  doprintf("cluster_size %u %u\n", fh1->cluster_size, fh2->cluster_size);
  if(sectors(fh1)!=0 || sectors(fh2)!=0)
    doprintf("sectors      %u %u\n", sectors(fh1), sectors(fh2));
  if(fh1->total_sect!=0 || fh2->total_sect!=0)
    doprintf("total_sect   %u %u\n", fh1->total_sect, fh2->total_sect);
  if(upart_type==UP_FAT32)
  {
    doprintf("fat32_length %u %u\n", fh1->fat32_length, fh2->fat32_length);
    doprintf("root_cluster %u %u\n", fh1->root_cluster, fh2->root_cluster);
  } else {
    doprintf("fat_length   %u %u\n", fh1->fat_length, fh2->fat_length);
    doprintf("dir_entries  %u %u\n", get_dir_entries(fh1), get_dir_entries(fh2));
  }
  return 0;
}

int dump_fat_info_rapport(const struct fat_boot_sector*fh1, const int upart_type)
{
  ecrit_rapport("cluster_size %u\n", fh1->cluster_size);
  if(sectors(fh1)!=0)
    ecrit_rapport("sectors      %u\n", sectors(fh1));
  if(fh1->total_sect!=0)
    ecrit_rapport("total_sect   %u\n", fh1->total_sect);
  if(upart_type==UP_FAT32)
  {
    ecrit_rapport("fat32_length %u\n", fh1->fat32_length);
    ecrit_rapport("root_cluster %u\n", fh1->root_cluster);
  } else {
    ecrit_rapport("fat_length   %u\n", fh1->fat_length);
    ecrit_rapport("dir_entries  %u\n", get_dir_entries(fh1));
  }
  return 0;
}

int dump_2fat_rapport(const struct fat_boot_sector*fh1, const struct fat_boot_sector*fh2, const int upart_type)
{
  switch(upart_type)
  {
    case UP_FAT12:
      ecrit_rapport("\nFAT12");
      break;
    case UP_FAT16:
      ecrit_rapport("\nFAT16");
      break;
    case UP_FAT32:
      ecrit_rapport("\nFAT32");
      break;
  }
  ecrit_rapport("\nsector_size  %u %u", sector_size(fh1),sector_size(fh2));
  ecrit_rapport("\ncluster_size %u %u", fh1->cluster_size,fh2->cluster_size);
  ecrit_rapport("\nreserved     %u %u", fh1->reserved,fh2->reserved);
  ecrit_rapport("\nfats         %u %u", fh1->fats,fh2->fats);
  ecrit_rapport("\ndir_entries  %u %u", get_dir_entries(fh1),get_dir_entries(fh2));
  ecrit_rapport("\nsectors      %u %u", sectors(fh1),sectors(fh2));
  ecrit_rapport("\nmedia        %02X %02X", fh1->media,fh2->media);
  ecrit_rapport("\nfat_length   %u %u", fh1->fat_length,fh2->fat_length);
  ecrit_rapport("\nsecs_track   %u %u", fh1->secs_track,fh2->secs_track);
  ecrit_rapport("\nheads        %u %u", fh1->heads,fh2->heads);
  ecrit_rapport("\nhidden       %u %u", fh1->hidden,fh2->hidden);
  ecrit_rapport("\ntotal_sect   %u %u", fh1->total_sect,fh2->total_sect);
  if(upart_type==UP_FAT32)
  {
	ecrit_rapport("\nfat32_length %u %u", fh1->fat32_length,fh2->fat32_length);
	ecrit_rapport("\nflags        %04X %04X", fh1->flags,fh2->flags);
	ecrit_rapport("\nversion      %u.%u  %u.%u", fh1->version[0], fh1->version[1],fh2->version[0], fh2->version[1]);
	ecrit_rapport("\nroot_cluster %u %u", fh1->root_cluster,fh2->root_cluster);
	ecrit_rapport("\ninfo_sector  %u %u", fh1->info_sector,fh2->info_sector);
	ecrit_rapport("\nbackup_boot  %u %u", fh1->backup_boot,fh2->backup_boot);
  }
  ecrit_rapport("\n");
  return 0;
}

int check_FAT(t_param_disk disk_car,t_diskext partition,const int debug)
{
  t_sector buffer;
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba)!=0)
  {
    ecrit_rapport(m_CHKFAT_RERR); return 1;
  }
  if(test_FAT(disk_car,(const struct fat_boot_sector *)buffer,partition,debug,0)!=0)
  {
    if(debug!=0)
    {
      ecrit_rapport("\n\ntest_FAT()\n");
      aff_part_rapport(2,disk_car,partition);
      dump_fat_rapport((const struct fat_boot_sector*)buffer, partition->upart_type);
    }
    return 1;
  }
  set_FAT_info(disk_car,(const struct fat_boot_sector *)buffer,partition,debug);
/*  aff_buffer(BUFFER_ADD,"Ok\n"); */
  return 0;
}

int set_FAT_info(t_param_disk disk_car, const struct fat_boot_sector *fat_header, t_diskext partition,const int debug)
{
  partition->name[0]=0;
  switch(partition->upart_type)
  {
    case UP_FAT12:
    case UP_FAT16:
      set_part_name(partition,((const char*)fat_header)+FAT1X_PART_NAME,11);
      break;
    case UP_FAT32:
      if(fat32_set_part_name(disk_car,partition,fat_header)==0)
      {
	if(debug!=0)
	  ecrit_rapport("set_FAT_info: name from BS used\n");
	set_part_name(partition,((const char*)fat_header)+FAT32_PART_NAME,11);
      }
      break;
  }
  return 0;
}

dword get_next_cluster(t_param_disk disk_car,const t_diskext partition, const int upart_type,const int offset, const dword cluster)
{
  /* Offset can be offset to FAT1 or to FAT2 */
  switch(upart_type)
  {
    case UP_FAT12:
      {
	unsigned char buffer[0x400];
	dword offset_s,offset_o;
	offset_s=(cluster+cluster/2)/512;
	offset_o=(cluster+cluster/2)%512;
	if(hd_read_LBA(disk_car,2, &buffer, partition->lba+offset+offset_s)!=0)
	{
	  ecrit_rapport("\nget_next_cluster error"); return 0;
	}
	if((cluster&1)!=0)
	  return (*((word*)&buffer[offset_o]))>>4;
	else
	  return (*((word*)&buffer[offset_o]))&0x0FFF;
      }
    case UP_FAT16:
      {
	unsigned char buffer[0x200];
	word *p16=(word*)&buffer;
	dword offset_s,offset_o;
	offset_s=cluster/256;
	offset_o=cluster%256;
	if(hd_read_LBA(disk_car,1, &buffer, partition->lba+offset+offset_s)!=0)
	{
	  ecrit_rapport("\nget_next_cluster error"); return 0;
	}
	return p16[offset_o];
      }
    case UP_FAT32:
      {
	unsigned char buffer[0x200];
	dword *p32=(dword*)&buffer;
	dword offset_s,offset_o;
	offset_s=cluster/128;
	offset_o=cluster%128;
	if(hd_read_LBA(disk_car,1, &buffer, partition->lba+offset+offset_s)!=0)
	{
	  ecrit_rapport("\nget_next_cluster error"); return 0;
	}
	/* FAT32 used 28 bits, the 4 high bits are reserved
	 * 0x00000000: free cluster
	 * 0x0FFFFFF7: bad cluster
	 * */

	return p32[offset_o]&0xFFFFFFF;
      }
  }
  return 0;
}

dword fat32_get_prev_cluster(t_param_disk disk_car,const t_diskext partition, const int fat_offset, const dword cluster, const dword no_of_cluster)
{
  unsigned char buffer[0x200];
  dword *p32=(dword*)&buffer;
  dword hd_offset=partition->lba+fat_offset;
  dword prev_cluster;
  for(prev_cluster=2;prev_cluster<=no_of_cluster;prev_cluster++)
  {
    unsigned int offset_s,offset_o;
    offset_s=prev_cluster/128;
    offset_o=prev_cluster%128;
    if((offset_o==0)||(prev_cluster==2))
    {
      if(hd_read_LBA(disk_car,1, &buffer, hd_offset++)!=0)
      {
	ecrit_rapport("fat32_get_prev_cluster error\n"); return 0;
      }
    }
    if((p32[offset_o] & 0xFFFFFFF) ==cluster)
      return prev_cluster;
  }
  return 0;
}

dword get_prev_cluster(t_param_disk disk_car,const t_diskext partition, const int upart_type,const int offset, const dword cluster, const dword no_of_cluster)
{
  dword prev_cluster;
  for(prev_cluster=2;prev_cluster<=no_of_cluster;prev_cluster++)
  {
    if(get_next_cluster(disk_car,partition,upart_type,offset, prev_cluster)==cluster)
      return prev_cluster;
  }
  return 0;
}

int test_FAT(t_param_disk disk_car,const struct fat_boot_sector *fat_header, t_diskext partition,const int debug, const int dump_ind)
{
  dword start_fat1,start_fat2,start_rootdir,start_data,no_of_cluster,fat_length,fat_length_calc,part_size,end_data;
  const char *buffer=(const char*)fat_header;
  if(fat_header->marker!=0xAA55)
  {
    aff_buffer(BUFFER_ADD,"test_FAT : " m_BAD_NMARK);
    ecrit_rapport("test_FAT : " m_BAD_NMARK);
    return 1;
  }
  if(dump_ind!=0)
    dump(stdscr,fat_header,0x200);
  if(!((fat_header->ignored[0]==0xeb&&fat_header->ignored[2]==0x90)||fat_header->ignored[0]==0xe9))
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_BAD_JUMP);
    ecrit_rapport(m_CHKFAT_BAD_JUMP);
    return 1;
  }
  if(sector_size(fat_header)!=512)
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_BYTESECT);
    ecrit_rapport(m_CHKFAT_BYTESECT);
    return 1;
  }
  switch(fat_header->cluster_size)
  { /* should depend of FAT12/16/32 */
    case 1:case 2:case 4:case 8:case 16:case 32:case 64:case 128: break;
    default:
      aff_buffer(BUFFER_ADD,m_CHKFAT_SECT_CLUSTER);
      ecrit_rapport(m_CHKFAT_SECT_CLUSTER);
      return 1;
  }
  if(fat_header->secs_track!=disk_car->CHS.sector)
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_SECTTRACK);
    ecrit_rapport("sect/track %d (FAT)!= %d (HD)\n",fat_header->secs_track,disk_car->CHS.sector);
    return 1;
  }
  if(fat_header->heads!=disk_car->CHS.head+1)
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_SIDE);
    ecrit_rapport("heads/cylinder %d (FAT) != %d (HD)\n",fat_header->heads,disk_car->CHS.head+1);
    return 1;
  }
  if(fat_header->fats!=2)
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_NBRFAT);
    ecrit_rapport(m_CHKFAT_NBRFAT);
    return 1;
  }
  if(get_dir_entries(fat_header)%16!=0)        /* Need to improved it */
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_ENTRY);
    ecrit_rapport(m_CHKFAT_ENTRY);
    return 1;
  }
  if(fat_header->media!=0xF8)
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_MEDIA);
    ecrit_rapport(m_CHKFAT_MEDIA);
    return 1;
  }

  if(memcmp(buffer+FAT_NAME1,"FAT12",5)==0) /* 2 Mo max */
  {
    part_size=sectors(fat_header);
    if(fat_header->reserved!=1)
    {
      aff_buffer(BUFFER_ADD,m_CHKFAT_RESERV);
      ecrit_rapport(m_CHKFAT_RESERV);
      return 1;
    }
    if(get_dir_entries(fat_header)==0)
    {
      aff_buffer(BUFFER_ADD,m_CHKFAT_ENTRY);
      ecrit_rapport(m_CHKFAT_ENTRY);
      return 1;
    }
    if(fat_header->fat_length>256)
    {
      aff_buffer(BUFFER_ADD,m_CHKFAT_SECTPFAT);
      ecrit_rapport(m_CHKFAT_SECTPFAT);
      return 1;
    }
    fat_length=fat_header->fat_length;
    start_fat1=fat_header->reserved;
    start_fat2=start_fat1+fat_length;
    start_rootdir=start_fat2+fat_length;
    start_data=start_rootdir+(get_dir_entries(fat_header)*32+512-1)/512;
    no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
    fat_length_calc=((no_of_cluster+1)*1.5/512);
    if((dword)((no_of_cluster+1)*1.5)%512!=0)
      fat_length_calc++;
    partition->upart_type=UP_FAT12;
  }
  else if(memcmp(buffer+FAT_NAME1,"FAT16",5)==0)
  {
    if(sectors(fat_header)!=0)
      part_size=sectors(fat_header);
    else
      part_size=fat_header->total_sect;
    fat_length=fat_header->fat_length;
    start_fat1=fat_header->reserved;
    start_fat2=start_fat1+fat_length;
    start_rootdir=start_fat2+fat_length;
    start_data=start_rootdir+(get_dir_entries(fat_header)*32+512-1)/512;
    no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
    fat_length_calc=((no_of_cluster+1)*2/512);
    if(((no_of_cluster+1)*2)%512!=0)
      fat_length_calc++;
    partition->upart_type=UP_FAT16;
  }
  else if(memcmp(buffer+FAT_NAME2,"FAT32",5)==0)
  {
    part_size=fat_header->total_sect;
    if(sectors(fat_header)!=0)
    {
      aff_buffer(BUFFER_ADD,m_CHKFAT_SIZE);
      ecrit_rapport(m_CHKFAT_SIZE);
      return 1;
    }
    fat_length=fat_header->fat_length>0?fat_header->fat_length:fat_header->fat32_length;
    start_fat1=fat_header->reserved;
    start_fat2=start_fat1+fat_length;
    start_data=start_fat2+fat_length;
    start_rootdir=start_data+(fat_header->root_cluster-2)*fat_header->cluster_size;
    no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
    fat_length_calc=((no_of_cluster+1)*4/512);
    if(((no_of_cluster+1)*4)%512!=0)
      fat_length_calc++;
    partition->upart_type=UP_FAT32;
  } else
  {
    ecrit_rapport("test_FAT unknown part_type\n");
    aff_buffer(BUFFER_ADD,"unknown part_type");
    return 1;
  }
  end_data=start_data+no_of_cluster*fat_header->cluster_size-1;
  if(partition->part_size>0)
  {
    if(part_size>partition->part_size)
    {
      ecrit_rapport("\ntest_FAT size boot_sector %ld > partition %ld\n",part_size,partition->part_size);
      aff_buffer(BUFFER_ADD,m_CHKFAT_SIZE);
      return 1;
    }
    else
    {
      if(debug && (part_size!=partition->part_size))
	ecrit_rapport("\ntest_FAT size boot_sector %ld, partition %ld\n",part_size,partition->part_size);
    }
  }
  if(debug>0)
  {
    ecrit_rapport("\n");
    aff_part_rapport(2,disk_car,partition);
    ecrit_rapport("FAT1 : %lu-%lu\n",start_fat1,start_fat1+fat_length-1);
    ecrit_rapport("FAT2 : %lu-%lu\n",start_fat2,start_fat2+fat_length-1);
    ecrit_rapport("start_rootdir : %lu",start_rootdir);
    if(partition->upart_type==UP_FAT32)
      ecrit_rapport(" root cluster : %u",fat_header->root_cluster);
    ecrit_rapport("\nData : %lu-%lu\n",start_data,end_data);
    ecrit_rapport("sectors : %lu\n",part_size);
    ecrit_rapport("cluster_size : %u\n",fat_header->cluster_size);
    ecrit_rapport("no_of_cluster : %lu (2 - %lu) => FAT", no_of_cluster,no_of_cluster+1);
    if(no_of_cluster<4085)
      ecrit_rapport("12");
    else
      if(no_of_cluster<65525)
	ecrit_rapport("16");
      else
	ecrit_rapport("32");
    ecrit_rapport("\nfat_length %lu calculated %lu\n",fat_length,fat_length_calc);
  }
  if(fat_length<fat_length_calc)
  { aff_buffer(BUFFER_ADD,m_CHKFAT_SECTPFAT); return 1; }
//    if(((partition->status==STATUS_LOG)&&(fat_header->hidden!=63)) || (fat_header->hidden!=partition->lba))
//  {	/* NEED RE-TESTING: mkdosfs has this little bug */
//    aff_buffer(BUFFER_ADD,m_CHKFAT_BAD_HIDDEN);
//    ecrit_rapport("Incorrect number of hidden sectors: fat_header->hidden=%d LBA=%d\n",fat_header->hidden,partition->lba);
//    return 1;
//  }
  return comp_FAT(disk_car,partition,fat_length,fat_header->reserved);
}

int comp_FAT(t_param_disk disk_car,const t_diskext partition,const dword taille,const dword sect_res)
{
  unsigned int reste=taille;
  unsigned int read_size;
  unsigned char buffer[NBR_SECT*0x200],buffer2[NBR_SECT*0x200];
  loff_t hd_offset, hd_offset2;
  hd_offset=partition->lba+sect_res;
  hd_offset2=hd_offset+taille;
  if(reste>1000) reste=1000;	/* Quick check ! */
  while(reste)
  {
    read_size=reste>NBR_SECT?NBR_SECT:reste;
    reste-=read_size;
    if(hd_read_LBA(disk_car,read_size, &buffer, hd_offset))
    { ecrit_rapport(m_CHKFAT_RERR);
      return 1;}
    if(hd_read_LBA(disk_car,read_size, &buffer2, hd_offset2))
    { ecrit_rapport(m_CHKFAT_RERR); 
      return 1;}
    if(memcmp(buffer,buffer2,0x200*read_size)!=0)
    { ecrit_rapport("FAT differs, FAT sector=%ld-%ld/%ld\n",taille-reste,taille-reste+read_size,taille); 
      return 1;
    }
    hd_offset+=read_size;
    hd_offset2+=read_size;
  }
  return 0;
}

unsigned int sector_size(const struct fat_boot_sector *fat_header)
{ return (fat_header->sector_size[1]<<8)+fat_header->sector_size[0]; }

unsigned int get_dir_entries(const struct fat_boot_sector *fat_header)
{ return (fat_header->dir_entries[1]<<8)+fat_header->dir_entries[0]; }

unsigned int sectors(const struct fat_boot_sector *fat_header)
{ return (fat_header->sectors[1]<<8)+fat_header->sectors[0]; }

int recover_FAT12(t_param_disk disk_car,const struct fat_boot_sector*fat_header, const t_diskext partition, const int debug, const int dump_ind)
{
  const char*buffer=(const char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME1,"FAT12",5)==0) /* 2 Mo max */
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nFAT12 at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(test_FAT(disk_car, fat_header, partition, debug, dump_ind))
	return 1;
      partition->part_size=sectors(fat_header);
      partition->part_type=P_12FAT;
//      partition->upart_type=UP_FAT12;	already done in test_FAT
      if(buffer[38]==0x29)	// BS_BootSig
	set_part_name(partition,buffer+FAT1X_PART_NAME,11);
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int recover_FAT16(t_param_disk disk_car,const struct fat_boot_sector*fat_header, t_diskext partition, const int debug, const int dump_ind)
{
  const char*buffer=(const char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME1,"FAT16",5)==0)
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nFAT16 at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(test_FAT(disk_car, fat_header, partition, debug, dump_ind))
	return 1;
      if(sectors(fat_header)!=0)
      {
	partition->part_size=sectors(fat_header);
	partition->part_type=P_16FAT;
      } else {
	partition->part_size=fat_header->total_sect;
	if(LBA2cylinder(disk_car,partition->lba+partition->part_size-1)<=1024)
	  partition->part_type=P_16FATBD;
	else
	  partition->part_type=P_16FATBD_LBA;
      }
//      partition->upart_type=UP_FAT16;	already done in test_FAT
      if(buffer[38]==0x29)	// BS_BootSig
	set_part_name(partition,buffer+FAT1X_PART_NAME,11);
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int recover_FAT32(t_param_disk disk_car, const struct fat_boot_sector*fat_header, t_diskext partition, const int debug, const int dump_ind,const int backup)
{
  const char*buffer=(const char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME2,"FAT32",5)==0)
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nFAT32 at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(test_FAT(disk_car, fat_header, partition, debug, dump_ind))
	return 1;
      if(sectors(fat_header)!=0)
	partition->part_size=sectors(fat_header);
      else
	partition->part_size=fat_header->total_sect;
      if(LBA2cylinder(disk_car,partition->lba+partition->part_size-1)<=1024)
	partition->part_type=P_32FAT;
      else
	partition->part_type=P_32FAT_LBA;
//      partition->upart_type=UP_FAT32;	already done in test_FAT
      if(backup)
      {
	partition->boot_sector=6;
	partition->lba-=6;	/* backup sector ... */
      }
      fat32_set_part_name(disk_car,partition,fat_header);
      return 0;
    }
  }     /* fin marqueur de fin =:-) */
  return 1;
}

int fat32_set_part_name(t_param_disk disk_car, t_diskext partition, const struct fat_boot_sector*fat_header)
{
  int i;
  unsigned char buffer[0x200*fat_header->cluster_size];
  if(hd_read_LBA(disk_car,fat_header->cluster_size, &buffer, partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat32_length+(fat_header->root_cluster-2)*fat_header->cluster_size))
  {
    ecrit_rapport(m_ROOT_CLUSTER_RERR); return 1;
  }
  for(i=0;(i<16*fat_header->cluster_size);i++)
  { /* Test attribut volume name and check if the volume name is erased or not */
    if(((buffer[i*0x20+0xB] & ATTR_EXT) !=ATTR_EXT) && ((buffer[i*0x20+0xB] & ATTR_VOLUME) !=0) && (buffer[i*0x20]!=0xE5))
    {
//      dump(stdscr,&buffer[i*0x20],0x20);
      set_part_name(partition,&buffer[i*0x20],11);
      return 2;
    }
  }
  return 0;
}

int recover_HPFS(t_param_disk disk_car,const struct fat_boot_sector*fat_header, t_diskext partition, const int debug, const int dump_ind)
{
  const char*buffer=(const char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+OS2_NAME,"IBM",3)==0)
    {   /* D'apres une analyse de OS2 sur systeme FAT...
	   FAT_NAME1=FAT
	 */
      if(debug||dump_ind)
      {
	ecrit_rapport("\nHPFS maybe at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(dump_ind)
	dump(stdscr,buffer,0x200);
      if(sectors(fat_header)!=0)
	partition->part_size=sectors(fat_header);
      else
	partition->part_size=fat_header->total_sect;
      partition->part_type=P_HPFS;
      partition->upart_type=UP_HPFS;
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int recover_OS2MB(t_param_disk disk_car, const struct fat_boot_sector*fat_header, t_diskext partition, const int debug, const int dump_ind)
{
  const char*buffer=(const char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME1,"FAT",3)==0)
    {   /* Attention a detecter apres OS2 */
      if(debug||dump_ind)
      {
	ecrit_rapport("\nMarker (0xAA55) at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(dump_ind)
	dump(stdscr,buffer,0x200);
      partition->part_size=(disk_car->CHS.head+1) * disk_car->CHS.sector; /* 1 cylinder */
      partition->part_type=P_OS2MB;
      partition->upart_type=UP_OS2MB;
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int is_fat(const int part_type)
{
  if(is_fat12(part_type)||is_fat16(part_type)||is_fat32(part_type))
    return 1;
  return 0;
}

int is_fat12(const int part_type)
{
  switch(part_type)
  {
    case P_12FAT:
    case P_12FATH:
      return 1;
  }
  return 0;
}

int is_fat16(const int part_type)
{
  switch(part_type)
  {
    case P_16FAT:
    case P_16FATH:
    case P_16FATBD:
    case P_16FATBDH:
      return 1;
  }
  return 0;
}

int is_fat32(const int part_type)
{
  switch(part_type)
  {
    case P_32FAT:
    case P_32FAT_LBA:
    case P_32FATH:
    case P_32FAT_LBAH:
      return 1;
  }
  return 0;
}
