/*

    File: fat_adv.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdio.h>
#include <ctype.h>	/* toupper */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "intrface.h"
#define MAX_TBFAT 10
int create_fat_boot_sector(t_param_disk disk_car,const t_diskext partition, const int reserved, const int debug, const unsigned int dir_entries, const dword root_cluster, const int cluster_size, const int fat_length,const int interface, const int dump_ind, const int upart_type);
t_fat_test start_fat[MAX_TBFAT];
int fat32_find_root_cluster(t_param_disk disk_car,const t_diskext partition,const int cluster_size, const dword no_of_cluster, const int offset,const int reserved,const int interface, const int debug);
int write_FAT_boot_code_aux(unsigned char *buffer);

/* The idea here is to be able to rebuild Boot Sectors and first FAT
 * when they've been erased by the CIH virus */
unsigned int get_frag(const unsigned char *buffer, const int fat_type)
{
  int nb_frag=1;
  switch(fat_type)
  {
    case 12:
      {
	unsigned int i=2;
	int low=0;
	int high=3;
	while(high<(0x200-1))
	{
	  unsigned int val;
	  if(low==0)
	    val=((buffer[high+1] & 0x0F) <<8) | buffer[high];
	  else
	    val=(buffer[high+1] <<4) | ((buffer[high]&0xF0)>>4);
//	  ecrit_rapport("get_frag FAT12 %X high=%u, low=%u - next=%X\n",i,high,low,val);
	  if(val!=0xFFF && val!=i+1)
	    nb_frag++;
	  if(low==0)
	    low=1;
	  else
	  {
	    high++;
	    low=0;
	  }
	  high++;
	  i++;
	}
      }
      break;
    case 16:
      {
	unsigned int i;
	const word *p16=(const word*)buffer;
	for(i=2; i<0x200/2-1; i++)
	{
//	  ecrit_rapport("get_frag FAT16 %X next=%X\n",i,p16[i]);
	  if(p16[i]!=0xFFFF && p16[i]!=i+1)
	    nb_frag++;
	}
      }
      break;
    case 32:
      {
	unsigned int i;
	const dword *p32=(const dword*)buffer;
      	for(i=2; i<0x200/4-1; i++)
	{
//	  ecrit_rapport("get_frag FAT32 %X next=%lX\n",i,p32[i]);
//      	  if(p32[i]!=(unsigned)0xFFFFFF0F && p32[i]!=i+1)
      	  if(p32[i]!=(unsigned)0x0FFFFFFF && p32[i]!=i+1)
	    nb_frag++;
	}
      }
      break;
  }
  return nb_frag;
}


int check_entree(const unsigned char *entree)
{ /* 1 si non nulle */
  int i;
  for(i=0;i<0x20;i++)
	if(*(entree+i))
	  return 1;
  return 0;
}

dword get_subdirectory(t_param_disk disk_car,const dword hd_offset,const dword i)
{
  unsigned char buffer[0x200];
  if(hd_read_LBA(disk_car,1, &buffer, hd_offset))
  {
	ecrit_rapport(m_ROOT_CLUSTER_RERR); return 1;
  }
//  dump(stdscr,buffer,0x200);
  /*                     12345678123*/
  if(strncmp(&buffer[0],".          ",8+3)!=0)
	return 1;
  if((unsigned)((buffer[0x15]<<24)|(buffer[0x14]<<16)|(buffer[0x1B]<<8)|buffer[0x1A])!=i)
	return 1;
  /*                        12345678123*/
  if(strncmp(&buffer[0x20],"..         ",8+3)!=0)
	return 1;
  return (buffer[0x35]<<24)+(buffer[0x34]<<16)+(buffer[0x3B]<<8)+buffer[0x3A];
}

int fat32_find_root_cluster(t_param_disk disk_car,const t_diskext partition,const int cluster_size, const dword no_of_cluster,const int offset, const int reserved,const int interface, const int debug)
{
  dword root_cluster;
  unsigned char buffer[cluster_size*0x200];
  int ind_stop=FALSE;
  if(interface)
  {
    move(22,0);
    standout();
    addstr("  Stop  ");
    standend();
  }
  for(root_cluster=2;(root_cluster<no_of_cluster)&&!ind_stop;root_cluster++)
  {
//    if(debug)
//      ecrit_rapport("fat32_find_root_cluster test cluster=%ld\n",root_cluster);
    if(interface)
    {
      move(9,0);
      clrtoeol();
      doprintf("Search root cluster %10ld/%ld",root_cluster,no_of_cluster);
      refresh();
    }
    /*
    */
    if(hd_read_LBA(disk_car,cluster_size, &buffer, partition->lba+offset+(root_cluster-2)*cluster_size))
    {
      ecrit_rapport(m_ROOT_CLUSTER_RERR); return 0;
    }
//    dump(stdscr,buffer,0x200);
    if(strncmp(&buffer[0],".          ",8+3)!=0)
    {
      int i,found=1;
      int etat=0,nb_subdir=0,nb_subdir_ok=0;
      for(i=0;found && (i<16*cluster_size);i++)
      {
//	if(debug)
//	  ecrit_rapport("fat32_find_root_cluster i=%d etat=%d\n",i,etat);
	if(check_entree(&buffer[i*0x20])==0)
	{
	  if(etat==0)
	    etat=1;
	}
	else
	  if(etat==1)
	  {
	    etat=2;
	    found=0;
	  }
	if((buffer[i*0x20]!=DELETED_FLAG) && (buffer[i*0x20+0xB]==0x10)) /* Test directory */
	{
	  nb_subdir++;
//	  ecrit_rapport("directory %s\n",&buffer[i*0x20]);
	}
      }
      if(nb_subdir<3)
      {
//	if(debug && nb_subdir)
//	  ecrit_rapport("Cluster %ld: not enough directory, only %d found\n",root_cluster,nb_subdir);
	found=0;
      }
      for(i=0;found && (i<16*cluster_size);i++)
      {
	if((buffer[i*0x20]!=DELETED_FLAG) && (buffer[i*0x20+0xB]==0x10)) /* Test directory */
	{
	  dword cluster=(buffer[i*0x20+0x15]<<24)+(buffer[i*0x20+0x14]<<16)+
	    (buffer[i*0x20+0x1B]<<8)+buffer[i*0x20+0x1A];
//	  ecrit_rapport("cluster %ld\n",cluster);
	  if((cluster>no_of_cluster)||(get_subdirectory(disk_car,partition->lba+offset+(cluster-2)*cluster_size,cluster)!=0))
	  {
//	    if(debug)
//	      ecrit_rapport("failed with %s\n",&buffer[i*0x20]);
//	    found=0;
	  }
	  else
	    nb_subdir_ok++;
	}
      }
//      ecrit_rapport("cluster %ld, etat=%d, found=%d,nb_subdir=%d,nb_subdir_ok=%d\n",root_cluster,etat,found,nb_subdir,nb_subdir_ok);
      if(found && (nb_subdir_ok>nb_subdir_ok*0.90))
      {
	dword tmp=root_cluster;
	int back=0;	/* To avoid an endless loop... */
	/* Il faut ajouter un parcours arriere de la FAT 
	 * car on localise le dernier cluster du root_cluster */
	do
	{
	  root_cluster=tmp;
	  tmp=fat32_get_prev_cluster(disk_car,partition,reserved,root_cluster,no_of_cluster);
	  ecrit_rapport("prev cluster %ld\n",root_cluster);
	} while(tmp && (++back<10));
	return root_cluster;
      }
    }
    if(interface)
    {
      switch(getch_nodelay())
      {
	case KEY_ENTER:
#ifdef DJGPP
      case PADENTER:
#endif
	case '\n':
	case '\r':
	case 's':
	case 'S':
	  ind_stop=1;
	  break;
      }
    }
  }
  ecrit_rapport("Search root cluster stopped: %10ld/%ld",root_cluster,no_of_cluster);
  return 0;
}

int analyse_dir_entries(t_param_disk disk_car,const t_diskext partition, const int offset,const int debug)
{
  unsigned char buffer[0x200];
  int i,j;
  int etat=0;
  dword hd_offset;
  hd_offset=partition->lba+offset;
  for(i=0;(i<200)&&(etat!=2);i++)
  {
    if(hd_read_LBA(disk_car,1, &buffer, hd_offset))
    {
      ecrit_rapport(m_ANALYSE_DIR_FAT_RERR); return 0;
    }
    for(j=0;j<16;j++)
    {
      if(check_entree(&buffer[j*0x20])==0)
      {
	if(etat==0)
	{
	  if(debug)
	    ecrit_rapport("dir_entries 0->1 %d\n",i*16+j);
	  etat=1;
	}
      }
      else
	if(etat==1)
	{
	  if(debug)
	    ecrit_rapport("dir_entries 1->2 %d\n",i*16+j);
	  etat=2;
	}
    }
    hd_offset++;
  }
  return (i-1)*16;
}

int create_fat_boot_sector(t_param_disk disk_car,const t_diskext partition, const int reserved, const int debug, const unsigned int dir_entries, const dword root_cluster, const int cluster_size, const int fat_length,const int interface, const int dump_ind, const int upart_type)
{
  unsigned char orgboot[3*0x200];
  unsigned char newboot[3*0x200];
  struct fat_boot_sector *org_fat_header=(struct fat_boot_sector *)&orgboot;
  struct fat_boot_sector *fat_header=(struct fat_boot_sector *)&newboot;
  int error=0;
  dword part_size=0;
  if(hd_read_LBA(disk_car,3, &orgboot, partition->lba))
  {
    ecrit_rapport(m_CHKFAT_RERR); return 1;
  }
  memcpy(&newboot,&orgboot,3*0x200);
  if(fat_header->marker!=0xAA55)
    write_FAT_boot_code_aux(newboot);
  fat_header->sector_size[0]=0x200 & 0xFF;
  fat_header->sector_size[1]=0x200>>8;
  fat_header->fats=2;
  fat_header->media=0xF8;
  fat_header->secs_track=disk_car->CHS.sector;
  fat_header->heads=disk_car->CHS.head+1;
  fat_header->marker=0xAA55;
  if(!((fat_header->ignored[0]==0xeb&&fat_header->ignored[2]==0x90)||fat_header->ignored[0]==0xe9))
  {
    fat_header->ignored[0]=0xeb;
    fat_header->ignored[2]=0x90;
  }

  /* I have seen a FAT32 partition that Win98 2nd edition was unable to read
   * because this name was missing! */
  if(strcmp(fat_header->system_id,"MSDOS5.0") &&
      strcmp(fat_header->system_id,"MSWIN4.0") &&
      strcmp(fat_header->system_id,"MSWIN4.1"))
    memcpy(fat_header->system_id,"MSWIN4.1",8);
  
  if(partition->status==STATUS_LOG)
    fat_header->hidden=63;
  else
    fat_header->hidden=partition->lba;
  fat_header->cluster_size=cluster_size;
  fat_header->reserved=reserved;
  /* The filesystem size can be smaller than the partition size */
  switch(upart_type)
  {
    case UP_FAT12:
      part_size=fat_header->reserved+2*fat_length+dir_entries*32/512+cluster_size*(fat_length*512*2/3-2);
      break;
    case UP_FAT16:
      part_size=fat_header->reserved+2*fat_length+dir_entries*32/512+cluster_size*(fat_length*512/2-2);
      break;
    case UP_FAT32:
      part_size=fat_header->reserved+2*fat_length+cluster_size*(fat_length*512/4-2);
      break;
    default:
      ecrit_rapport("create_fat_boot_sector severe error\n");
      exit(1);
  }
  if(part_size>partition->part_size)
    part_size=partition->part_size;
  if(part_size>0xFFFF)
  {
    fat_header->sectors[0]=0;
    fat_header->sectors[1]=0;
    fat_header->total_sect=part_size;
  }
  else
  {
    fat_header->sectors[1]=part_size>>8;
    fat_header->sectors[0]=part_size;
    fat_header->total_sect=0;
  }
  switch(upart_type)
  {
    case UP_FAT12:
      if((fat_length==0) || (dir_entries==0))
	error=1;
      if((newboot[36]<0x80)||(newboot[36]>0x88))
	newboot[36]=0x80; // BS_DrvNum=0x80
      newboot[37]=0;	// BS_Reserved1=0
      newboot[38]=0x29; 	// Boot sig=0x29
      memcpy(newboot+FAT_NAME1,"FAT12",5);
      if(memcmp(newboot+FAT_NAME2,"FAT32",5)==0)
	memcpy(newboot+FAT_NAME2, "        ",8);
      fat_header->fat_length=fat_length;
      fat_header->dir_entries[1]=dir_entries>>8;
      fat_header->dir_entries[0]=dir_entries;
      break;
    case UP_FAT16:
      if((fat_length==0) || (dir_entries==0))
	error=1;
      if((newboot[36]<0x80)||(newboot[36]>0x88))
	newboot[36]=0x80; // BS_DrvNum
      newboot[37]=0;	// BS_Reserved1=0
      newboot[38]=0x29; 	// Boot sig=0x29
      memcpy(newboot+FAT_NAME1,"FAT16",5);
      if(memcmp(newboot+FAT_NAME2,"FAT32",5)==0)
	memcpy(newboot+FAT_NAME2, "        ",8);
      fat_header->fat_length=fat_length;
      fat_header->dir_entries[1]=dir_entries>>8;
      fat_header->dir_entries[0]=dir_entries;
      break;
    case UP_FAT32:
      if((fat_length==0) || (root_cluster==0))
	error=1;
      /*
	 Bits 0-3 -- Zero-based number of active FAT. Only valid if mirroring
	 is disabled.
	 Bits 4-6 -- Reserved.
	 Bit    7 -- 0 means the FAT is mirrored at runtime into all FATs.
	 -- 1 means only one FAT is active; it is the one referenced
	 in bits 0-3.
	 Bits 8-15 -- Reserved.
       */
      fat_header->flags=0;
      fat_header->version[0]=0;
      fat_header->version[1]=0;
      fat_header->fat_length=0;
      fat_header->dir_entries[0]=0;
      fat_header->dir_entries[1]=0;
      fat_header->fat32_length=fat_length;
      /* Sector number of FSINFO structure in the
       * reserved area of the FAT32 volume.
       */
      fat_header->info_sector=1;
      fat_header->backup_boot=6;
      fat_header->root_cluster=root_cluster;
      memset(&fat_header->reserved2,0,sizeof(fat_header->reserved2));
      if((newboot[64]<0x80)||(newboot[64]>0x88))
	newboot[64]=0x80; // BS_DrvNum=0x80
      newboot[65]=0;	// BS_Reserved1=0
      newboot[66]=0x29; 	// Boot sig=0x29
      if((memcmp(newboot+FAT_NAME1,"FAT12",5)==0) ||(memcmp(newboot+FAT_NAME1,"FAT16",5)==0))
	memcpy(newboot+FAT_NAME1,"        ",8);
      memcpy(newboot+FAT_NAME2,  "FAT32   ",8);
      newboot[0x1FC]=0x00;	/* part of the signature */
      newboot[0x1FD]=0x00;
      memset(&newboot[0x200],0,0x200);
      newboot[0x200]='R';		/* Signature RRaA */
      newboot[0x201]='R';
      newboot[0x202]='a';
      newboot[0x203]='A';
      newboot[0x3E4]='r';		/* Signature rrAa */
      newboot[0x3E5]='r';
      newboot[0x3E6]='A';
      newboot[0x3E7]='a';
      /* Don't compute the number of free cluster or the next free cluster */
      newboot[0x3E8]=0xFF;	/* Free clusters on disk */
      newboot[0x3E9]=0xFF;
      newboot[0x3EA]=0xFF;
      newboot[0x3EB]=0xFF;
      newboot[0x3EC]=0xFF;	/* Next avaible clusters */
      newboot[0x3ED]=0xFF;
      newboot[0x3EE]=0xFF;
      newboot[0x3EF]=0xFF;
      newboot[0x3FC]=0x00;	/* End of Sector signature */
      newboot[0x3FD]=0x00;
      newboot[0x3FE]=0x55;
      newboot[0x3FF]=0xAA;
      newboot[0x5FC]=0x00;	/* End of Sector signature */
      newboot[0x5FD]=0x00;
      newboot[0x5FE]=0x55;
      newboot[0x5FF]=0xAA;
      break;
  }
  if(memcmp(newboot,orgboot,1*0x200))	/* Only compare the first sector */
  {
    ecrit_rapport("             New / Current boot sector");
    dump_2fat_rapport(fat_header,org_fat_header,upart_type);
    ecrit_rapport("Extrapolated boot sector and current boot sector are different.\n");
  }
  else
  {
    ecrit_rapport("Extrapolated boot sector and current boot sector are identical.\n");
  }
  if(error)
    ecrit_rapport("Bad extrapolation.\n");
  /* */
  if(interface)
  {
    struct MenuItem menuSaveBoot[]=
    {
      { 'Q',"Quit","Quit this section"},
      { 'D', "Dump", "Dump sector" },
      { 'W', "Write","Write boot"},
      { 0, NULL, NULL }
    };
    const char *options="DQ";
    int do_write=0;
    int do_dump=0;
    do
    {
      do_dump=0;
      aff_copy(stdscr);
      move(4,0);
      aff_disk(stdscr,disk_car);
      mvaddstr(6,0,m_PART_HEADER2);
      move(7,0);
      aff_part(stdscr,2,disk_car,partition);
      move(8,0);
      if(memcmp(newboot,orgboot,0x200))	/* Only compare the first sector */
      {
	options="WDQ";
	dump_2fat_info(fat_header, org_fat_header, upart_type);
	doprintf("Extrapolated boot sector and current boot sector are different.\n");
	if(error)
	  ecrit_rapport("Warning: Extrapolated boot sector have incorrect values.\n");
      }
      else
      {
	dump_fat_info(fat_header, upart_type);
	doprintf("Extrapolated boot sector and current boot sector are identical.\n");
      }
      switch(toupper(menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuSaveBoot,8,options,MENU_HORIZ | MENU_ACCEPT_OTHERS, 0)))
      {
	case 'W':
	  if(strchr(options,'W'))
	    do_write=1;
	  break;
	case 'D':
	  if(strchr(options,'D'))
	  {
	    do_dump=1;
	    dump2(stdscr,newboot,orgboot, upart_type==UP_FAT32?3*0x200:0x200);
	  }
	  break;
      }
    } while(!do_write && do_dump);
    if(do_write && ask_confirmation("Confirm ? (Y/N)"))
    {
      ecrit_rapport("Write new boot!\n");
      /* Write boot sector and backup boot sector */
      if(upart_type==UP_FAT32)
      {
	hd_write_LBA(disk_car,3, &newboot, partition->lba);
	hd_write_LBA(disk_car,3, &newboot, partition->lba+fat_header->backup_boot);
      }
      else
	hd_write_LBA(disk_car,1, &newboot, partition->lba);
      /* Use another utility to check and correct the FAT */
    }
    else
      ecrit_rapport("Don't write new boot!\n");
  }
  return 0;
}

int calcul_cluster_size(const int upart_type, const dword data_size, const int fat_length)
{
//  ecrit_rapport("calcul_cluster_size data_size=%lu, fat_length=%u\n",data_size,fat_length);
  switch(upart_type)
  {
    case UP_FAT12:
      return up2power(data_size/(fat_length*512/1.5-1));
    case UP_FAT16:
      return up2power(data_size/(fat_length*512/2-1));
    case UP_FAT32:
      return up2power(data_size/(fat_length*512/4-1));
  }
  return 0;
}

int search_fat_aux(t_param_disk disk_car,const t_diskext partition,const dword max_offset,const int p_fat12,const int p_fat16,const int p_fat32,int *f_fat12,int*f_fat16,int*f_fat32,const int debug,const int dump_ind,const int interface)
{
  dword offset;
  int pos_fat_test=0;
  dword hd_offset;
  if(debug)
	ecrit_rapport("search_fat_aux(partition,max_offset=%ld,p_fat12=%d,p_fat16=%d,p_fat32=%d,*f_fat12,*f_fat16,*f_fat32,debug=%d,dump_ind=%d)\n",max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);
  for(offset=0,hd_offset=partition->lba;(offset<max_offset)&&(pos_fat_test<MAX_TBFAT);offset++,hd_offset++)
  {
    unsigned char buffer[0x200];
    int inc_ind=0;
    if(interface && (offset%100==0))
    {
      move(8,30);
      clrtoeol();	/* before addstr for BSD compatibility */
      doprintf("analyse %10ld/%ld",offset,max_offset);
      refresh();
    }
    if(hd_read_LBA(disk_car,1, &buffer, hd_offset))
    {
      continue;
    }
    /* F8 FF FF FF : FAT 16
     * F8 FF FF 7F
     * F8 FF FF FF FF FF FF FF : FAT 32 Windows is not running
     * F8 FF FF 0F FF FF FF FF : FAT 32
     * F8 FF FF FF FF FF FF F7 : FAT32	Windows is running
     *					Shutdown Flag is 'cleared')
     *					*/
    if(buffer[0]==0xF8 && buffer[1]==0xFF)
    {
      start_fat[pos_fat_test].offset=offset;
      start_fat[pos_fat_test].f_fat12=0;
      start_fat[pos_fat_test].f_fat16=0;
      start_fat[pos_fat_test].f_fat32=0;
      if(p_fat32==1 && buffer[2]==0xFF && ((buffer[3]==0x0F) ||(buffer[3]==0xFF)) && buffer[4]==0xFF && buffer[5]==0xFF && buffer[6]==0xFF)
      {
	start_fat[pos_fat_test].f_fat32=get_frag(buffer,32);
	if(*f_fat32==0 || start_fat[pos_fat_test].f_fat32<*f_fat32)
	  *f_fat32=start_fat[pos_fat_test].f_fat32;
	inc_ind=1;
      }
      if(p_fat16==1 && buffer[2]==0xFF && ((buffer[3] & 0xF7)==0xF7))
      {
	start_fat[pos_fat_test].f_fat16=get_frag(buffer,16);
	if(*f_fat16==0 || start_fat[pos_fat_test].f_fat16<*f_fat16)
	  *f_fat16=start_fat[pos_fat_test].f_fat16;
	inc_ind=1;
      }
      if(p_fat12==1)
      {
	start_fat[pos_fat_test].f_fat12=get_frag(buffer,12);
	if(*f_fat12==0 || start_fat[pos_fat_test].f_fat12<*f_fat12)
	  *f_fat12=start_fat[pos_fat_test].f_fat12;
	inc_ind=1;
      }
      if(inc_ind)
      {
	if(dump_ind)
	  dump(stdscr,buffer,0x200);
	if(debug)
	{
	  ecrit_rapport("%5ld  12:%d 16:%d 32:%d\n",start_fat[pos_fat_test].offset,start_fat[pos_fat_test].f_fat12,start_fat[pos_fat_test].f_fat16,start_fat[pos_fat_test].f_fat32);
	  ecrit_rapport("best  12:%d 16:%d 32:%d\n",*f_fat12,*f_fat16,*f_fat32);
	}
	pos_fat_test++;
      }
    }
  }
  return pos_fat_test;
}

int get_part_type(const int f_fat12, const int f_fat16, const int f_fat32, const int debug)
{
  int upart_type=0;
  int f_part_type=0;
  if(debug)
    ecrit_rapport("get_part_type(f_fat12=%d,f_fat16=%d,f_fat32=%d,debug=%d) => ",f_fat12,f_fat16,f_fat32,debug);
  if(f_fat12)
  {
    upart_type=UP_FAT12;
    f_part_type=f_fat12;
  }
  if(f_fat16 &&(!f_part_type ||(f_fat16<f_part_type)))
  {
    upart_type=UP_FAT16;
    f_part_type=f_fat16;
  }
  if(f_fat32 &&(!f_part_type ||(f_fat32<f_part_type)))
  {
    upart_type=UP_FAT32;
    f_part_type=f_fat32;
  }
  if(debug)
  {
    switch(upart_type)
    {
      case UP_FAT12: ecrit_rapport("FAT12\n"); break;
      case UP_FAT16: ecrit_rapport("FAT16\n"); break;
      case UP_FAT32: ecrit_rapport("FAT32\n"); break;
      default: ecrit_rapport("Not a FAT\n"); break;
    }
  }
  return upart_type;
}

int select_fat(const int pos_fat_test, const int upart_type, const int f_fat12, const int f_fat16, const int f_fat32, const int debug)
{
  int i,pos=0;
  switch(upart_type)
  {
	case UP_FAT12:
	  for(i=0;i<pos_fat_test;i++)
	  {
		if(start_fat[i].f_fat12==f_fat12)
		{
		  start_fat[pos].offset=start_fat[i].offset;
		  start_fat[pos].f_fat12=start_fat[i].f_fat12;
		  if(debug)
			ecrit_rapport("FAT12 at %lu\n",start_fat[pos].offset);
		  pos++;
		}
	  }
	  break;
	case UP_FAT16:
	  for(i=0;i<pos_fat_test;i++)
	  {
		if(start_fat[i].f_fat16==f_fat16)
		{
		  start_fat[pos].offset=start_fat[i].offset;
		  start_fat[pos].f_fat16=start_fat[i].f_fat16;
		  if(debug)
			ecrit_rapport("FAT16 at %lu\n",start_fat[pos].offset);
		  pos++;
		}
	  }
	  break;
	case UP_FAT32:
	  for(i=0;i<pos_fat_test;i++)
	  {
		if(start_fat[i].f_fat32==f_fat32)
		{
		  start_fat[pos].offset=start_fat[i].offset;
		  start_fat[pos].f_fat32=start_fat[i].f_fat32;
		  if(debug)
			ecrit_rapport("FAT32 at %lu\n",start_fat[pos].offset);
		  pos++;
		}
	  }
	  break;
  }
  return pos;
}

int find_fat_type(t_param_disk disk_car,int*reserved, int*fat_length, const t_diskext partition,const dword max_offset,const int p_fat12,const int p_fat16,const int p_fat32,const int debug,const int dump_ind,const int interface)
{
  int pos_fat_test=0;
  int f_fat12=0, f_fat16=0, f_fat32=0;
  int upart_type;
#ifdef TESTING
  if(debug)
	ecrit_rapport("find_fat_type(*reserved,*fat_length,partition,max_offset=%d,p_fat12=%d,p_fat16=%d,p_fat32=%d,debug=%d,dump_ind=%d)\n",max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);
#endif
  pos_fat_test=search_fat_aux(disk_car,partition,max_offset,p_fat12,p_fat16,p_fat32,&f_fat12,&f_fat16,&f_fat32,(debug>0),dump_ind,interface);
#ifdef TESTING
  ecrit_rapport("pos_fat_test=%d\n",pos_fat_test);
#endif
  if(debug)
    ecrit_rapport("search_fat_aux result 12:%d 16:%d 32:%d\n",f_fat12,f_fat16,f_fat32);
  upart_type=get_part_type(1.5*f_fat12, 2*f_fat16, 4*f_fat32,debug);
  /* Run again search_fat_aux */
  f_fat12=0; f_fat16=0; f_fat32=0;
  switch(upart_type)
  {
	case UP_FAT12:
	  *reserved=1;
	  pos_fat_test=search_fat_aux(disk_car,partition,max_offset,1,0,0,&f_fat12,&f_fat16,&f_fat32,(debug>1),0,interface);
	  break;
	case UP_FAT16:
	  *reserved=1;
	  pos_fat_test=search_fat_aux(disk_car,partition,max_offset,0,1,0,&f_fat12,&f_fat16,&f_fat32,(debug>1),0,interface);
	  break;
	case UP_FAT32:
	  *reserved=32;
	  pos_fat_test=search_fat_aux(disk_car,partition,max_offset,0,0,1,&f_fat12,&f_fat16,&f_fat32,(debug>1),0,interface);
	  break;
	default:
	  ecrit_rapport("find_fat_type failed\n");
	  return 0;
  }

  pos_fat_test=select_fat(pos_fat_test,upart_type,f_fat12,f_fat16,f_fat32,debug);
//  ecrit_rapport("FAT at %ld %ld, pos_fat_test=%d\n",start_fat[0].offset,start_fat[1].offset,pos_fat_test);
  switch(pos_fat_test)
  {
	case 0:
	  *fat_length=0;
	  break;
	case 1:
	  /* Idea: the second FAT is still present, not the first */
	  *fat_length=start_fat[0].offset-*reserved;
	  break;
	default:
	  *reserved=start_fat[0].offset;
	  *fat_length=start_fat[1].offset-*reserved;
	  break;
  }
  return upart_type;
}

int rebuild_FAT_BS(t_param_disk disk_car, t_diskext partition, const int debug, const int dump_ind,const int interface)
{
  unsigned int fat_length;
  dword fat_length_max;
  int p_fat12,p_fat16,p_fat32;
  unsigned int cluster_size_min=1;
  unsigned int cluster_size;
  dword max_offset;
  dword root_cluster=0;
  dword start_data,data_size;
  int upart_type;
  unsigned int dir_entries=0;
  unsigned int reserved;
/*
 * dump ???
 *
 * */
  if(partition->part_size>2*1024*1024*2+63)
  {
    p_fat32=1;
    p_fat16=0;
    p_fat12=0;
  }
  else
/* 1<<12 clusters * 8 secteurs/clusters= 32768 secteurs 
   fat_length=((1<<12+1)*1.5/512)+1=13; */
    if(partition->part_size>=(1+2*13+32768+63))
    {
      p_fat32=1;
      p_fat16=1;
      p_fat12=0;
    }
    else
    {
      p_fat32=0;
      p_fat16=1;
      p_fat12=1;
    }
#ifdef TESTING
  p_fat32=1; p_fat16=1; p_fat12=1;
#endif
  if(debug)
  {
    aff_part_rapport(1,disk_car,partition);
    ecrit_rapport("\nrebuild_FAT_BS p_fat12 %d, p_fat16 %d, p_fat32 %d\n", p_fat12,p_fat16,p_fat32);
  }
  /* Set fat_length_max */
  if(p_fat32)
  {	/* Cluster 1k */
    fat_length_max=partition->part_size/cluster_size_min*4/512;
  }
  else
  if(p_fat16)
  {
    while(partition->part_size/cluster_size_min>(1<<16))
      cluster_size_min*=2;
    fat_length_max=partition->part_size/cluster_size_min*2/512;
  }
  else
  {
    while(partition->part_size/cluster_size_min>(1<<12))
      cluster_size_min*=2;
    fat_length_max=partition->part_size/cluster_size_min*1.5/512;
  }
  if(debug)
  {
    ecrit_rapport("cluster_size_min %d\n",cluster_size_min);
    ecrit_rapport("fat_length_max %ld\n", fat_length_max);
  }
  max_offset=fat_length_max+64;
  /*
  if(debug)
    ecrit_rapport("search_fat16(partition,max_offset=%d,p_fat12=%d,p_fat16=%d,p_fat32=%d,debug=%d,dump_ind=%d)\n",max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);
    */
  if(interface)
  {
    aff_copy(stdscr);
    move(4,0);
    aff_disk(stdscr,disk_car);
    mvaddstr(6,0,m_PART_HEADER2);
    move(7,0);
    aff_part(stdscr,2,disk_car,partition);
    move(8,0);
    doprintf("FAT : %s %s %s?\n",p_fat12?"12 ":"", p_fat16?"16 ":"", p_fat32?"32 ":"");
    refresh();
  }
  upart_type=find_fat_type(disk_car,&reserved, &fat_length, partition,max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind,interface);
  if(interface)
  {
    move(8,0);
    clrtoeol();
    switch(upart_type)
    {
      case UP_FAT12:
	addstr("FAT : 12");
	break;
      case UP_FAT16:
	addstr("FAT : 16");
	break;
      case UP_FAT32:
	addstr("FAT : 32");
	break;
      default:
	addstr("Not a FAT");
	return 0;
    }
    if(fat_length==0)
      addstr(" Can't find FAT length\n");
    refresh();
  }
  if(upart_type && (fat_length>0))
  {
    start_data=reserved+2*fat_length;
    /* FAT1x: Find size of root directory */
    if((upart_type==UP_FAT12) || (upart_type==UP_FAT16))
    {
      dir_entries=analyse_dir_entries(disk_car,partition,reserved+2*fat_length,debug);
      start_data+=(dir_entries-1)/16+1;
    }
    if(partition->part_size<start_data)
    {
      ecrit_rapport("Error part_size=%lu < start_data=%lu\n",partition->part_size,start_data);
    }
    data_size=partition->part_size-start_data;
    /* Get Cluster Size */
    cluster_size=calcul_cluster_size(upart_type,data_size,fat_length);
    if(debug)
      ecrit_rapport("cluster_size %d\n",cluster_size);
    /* FAT32 : Find root cluster */
    if(upart_type==UP_FAT32)
      root_cluster=fat32_find_root_cluster(disk_car,partition,cluster_size,data_size/cluster_size,start_data,reserved,interface,debug);
    if(interface)
    {
      move(9,0);
      clrtoeol();
      refresh();
    }
    create_fat_boot_sector(disk_car,partition, reserved, debug,dir_entries,root_cluster,cluster_size,fat_length,interface,dump_ind,upart_type);
    if(debug)
    {
      aff_part_rapport(1,disk_car,partition);
      ecrit_rapport("\n");
    }
  }
  return 0;
}

int write_FAT_boot_code_aux(unsigned char *buffer)
{
  const unsigned char boot_code[0x200]= {
    0xeb, 0x3c, 0x90, 0x6d, 0x6b, 0x64, 0x6f, 0x73, 0x66, 0x73, 0x00, 0x00, 0x02, 0x08, 0x01, 0x00,
    0x02, 0x00, 0x02, 0x00, 0x00, 0xf8, 0xcc, 0x00, 0x3f, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x5a, 0x5f, 0x06, 0x00, 0x00, 0x00, 0x29, 0xf8, 0x3f, 0x7c, 0x3e, 'T',   'E',   'S', 'T',  'D',
     'I',  'S',  'K', 0x20, 0x20, 0x20, 0x46, 0x41, 0x54, 0x31, 0x36, 0x20, 0x20, 0x20, 0x0e, 0x1f,
    0xbe, 0x5b, 0x7c, 0xac, 0x22, 0xc0, 0x74, 0x0b, 0x56, 0xb4, 0x0e, 0xbb, 0x07, 0x00, 0xcd, 0x10,
    0x5e, 0xeb, 0xf0, 0x32, 0xe4, 0xcd, 0x16, 0xcd, 0x19, 0xeb, 0xfe,  'T', 'h',   'i',  's',  ' ',
     'i',  's',  ' ',  'n',  'o',  't',  ' ',  'a',  ' ',  'b',  'o',  'o',  't',  'a',  'b',  'l',
     'e',  ' ',  'd',  'i',  's',  'k',  '.',  ' ',  ' ',  'P',  'l',  'e',  'a',  's',  'e',  ' ',
     'i',  'n',  's',  'e',  'r',  't',  ' ',  'a',  ' ',  'b',  'o',  'o',  't',  'a',  'b',  'l',
     'e',  ' ',  'f',  'l',  'o',  'p',  'p',  'y',  ' ',  'a',  'n',  'd', 0x0d, 0x0a,  'p',  'r',
     'e',  's',  's',  ' ',  'a',  'n',  'y',  ' ',  'k',  'e',  'y',  ' ',  't',  'o',  ' ',  't',
     'r', 'y',   ' ',  'a',  'g',  'a',  'i',  'n',  ' ',  '.',  '.',  '.',  ' ', 0x0d, 0x0a, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x55, 0xaa
  };
  memcpy(buffer,&boot_code,0x200);
  return 0;
}
