/*

    File: godmode.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#define DEBUG 1
#include <ctype.h>      /* toupper, tolower */
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "analyse.h"
#include "lang.h"
#include "godmode.h"
#include "testdisk.h"
#include "fat.h"
#include "intrface.h"
#include "hdaccess.h"
#include "md.h"
//#define ESC 0x1B
#define RO 1
#define RW 0
extern char* nom_os[256];
void align_structure(t_param_disk disk_car, t_list_part list_part);
void init_structure(t_param_disk disk_car,t_list_part list_part);
void init_part_order(t_list_part list_part);
void ask_mbr_order(t_param_disk disk_car,t_list_part list_part);
t_list_part reduce_structure(t_list_part list_part);
t_list_part add_ext_part(t_param_disk disk_car, t_list_part list_part, const int max_ext);
int write_mbr(t_param_disk disk_car, t_list_part list_part,int,int);
int write_all_log(t_param_disk disk_car, t_list_part list_part,int,int);
void adjust_part_size(t_param_disk,t_diskext);
int diff(unsigned char buffer[0x200], unsigned char buffer_org[0x200]);
t_list_part do_recover_cih(t_param_disk disk_car,t_list_part list_part,int debug, const int dump_ind);

void align_structure(t_param_disk disk_car, t_list_part list_part)
{
  t_list_part element;
  for(element=list_part;element;element=element->next)
  {
    t_CHS end;
    LBA2CHS(disk_car,element->part->lba+element->part->part_size-1,&end);
    end.sector=disk_car->CHS.sector;
    end.head=disk_car->CHS.head;
    element->part->part_size=CHS2LBA(disk_car, &end)-element->part->lba+1;
  }
}

void init_structure(t_param_disk disk_car,t_list_part list_part)
{
  /* Reconstruit une structure 
   * nbr_prim nombre de partition primaire
   * end_log_block : partition etendu du block etendu
   * nbr_log_block : nombre de block de partition etendu
   *                    devrait etre 0 ou 1 */
  int nbr_prim=0, nbr_log_block=0;
  t_list_part element;
  t_list_part end_log_block=NULL,end_biggest_log_block=NULL;
  int log_block_size=0,biggest_log_block_size=0;
  t_list_part new_list_part=NULL;
//  for(element=list_part;element;element=element->next)
//    ecrit_rapport("%p %p %p %d\n",element,element->prev,element->next,element->part->status);
  /* Create new list */
  for(element=list_part;element;element=element->next)
    element->to_be_removed=0;
  for(element=list_part;element;element=element->next)
  {
    int to_be_removed=0;
    t_list_part element2;
    for(element2=element->next;element2;element2=element2->next)
      if(element->part->lba+element->part->part_size-1 >= element2->part->lba)
      {
	to_be_removed=1;
	element2->to_be_removed=1;
      }
    if(to_be_removed)
      element->to_be_removed=1;
    if(!element->to_be_removed)
      new_list_part=insert_new_partition(new_list_part,element->part);
  }
  /* Verify */
  for(element=new_list_part;element;element=element->next)
  {
    if(!can_be_ext(disk_car,element->part))
    {
      nbr_prim++;
      if(end_log_block && (end_log_block->next==element))
      {
	if(log_block_size>biggest_log_block_size)
	{
	  biggest_log_block_size=log_block_size;
	  end_biggest_log_block=end_log_block;
	}
	nbr_log_block++;
	end_log_block=NULL;
      }
    }
    else
    {
      log_block_size++;
      end_log_block=element;
    }
  }
  /* Verification */
  if(end_log_block && (end_log_block->next==NULL))
  {
    if(log_block_size>biggest_log_block_size)
    {
      biggest_log_block_size=log_block_size;
      end_biggest_log_block=end_log_block;
    }
    nbr_log_block++;
  }
  ecrit_rapport("\nRes: nbr_prim %d, nbr_log_block %d\n", nbr_prim, nbr_log_block);
  /* Set primary, extended, logical */
  if(nbr_prim+nbr_log_block<=4)
  {
    int set_prim_bootable_done=0;
    for(element=end_biggest_log_block;element&&can_be_ext(disk_car,element->part);element=element->prev)
    {
//      ecrit_rapport("%p %p %p %d\n",element,element->prev,element->next,element->part->status);
      element->part->status=STATUS_LOG;
    }
    for(element=new_list_part;element;element=element->next)
    {
      if(element->part->status!=STATUS_LOG)
      {
	if(set_prim_bootable_done)
	  element->part->status=STATUS_PRIM;
	else
	{
	  /* The first primary partition is bootable unless it's a swap */
	  if(element->part->upart_type!=UP_LINSWAP)
	  {
	    element->part->status=STATUS_PRIM_BOOT;
	    set_prim_bootable_done=1;
	  }
	}
      }
    }
  }
  check_list_part(new_list_part);
  if(test_structure(new_list_part))
  {
    for(element=new_list_part;element;element=element->next)
      element->part->status=STATUS_DELETED;
  }
  {
    /* free */
    t_list_part new_element;
    for(element=new_list_part;element;element=new_element)
    {
      new_element=element->next;
      FREE(element);
    }
  }
  check_list_part(list_part);
}

int can_be_ext(const t_param_disk disk_car, t_diskext partition)
{
  return((LBA2head(disk_car,partition->lba)>0)&&
      (LBA2cylinder(disk_car,partition->lba)!=0 ||
       LBA2head(disk_car,partition->lba)!=1 ||
       LBA2sector(disk_car,partition->lba)!=1));
}

void only_one_bootable( t_list_part list_part, t_list_part part_boot)
{
  t_list_part element;
  if(part_boot->part->status==STATUS_PRIM_BOOT)
    for(element=list_part;element;element=element->next)
    {
      if((element!=part_boot)&&(element->part->status==STATUS_PRIM_BOOT))
	element->part->status=STATUS_PRIM;
    }
}

int test_structure(t_list_part list_part)
{ /* Return 1 if bad*/
  int nbr_prim=0, nbr_prim_boot=0, nbr_log_block=0;
  t_list_part first_log=NULL;
  t_list_part new_list_part=NULL,element,new_element;
  int res=0;
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
    {
      case STATUS_LOG:
	  if(first_log==NULL)
	  {
	    first_log=element;
	    nbr_log_block++;
	  }
	break;
      case STATUS_PRIM_BOOT:
	if(nbr_prim_boot++)
	  return 1;
      case STATUS_PRIM:
	nbr_prim++;
	first_log=NULL;
	break;
      case STATUS_DELETED:
	break;
    }
  }
  if(nbr_log_block>1 || nbr_log_block+nbr_prim>4)
    return 1;
  /* Sort list_part in new_list_part */
  for(element=list_part;element;element=element->next)
  {
    if(element->part->status!=STATUS_DELETED)
      new_list_part=insert_new_partition(new_list_part,element->part);
  }
  /* Test chevauchement */
//  ecrit_rapport("\nTest chevauchement\n");
  for(element=new_list_part;element;element=element->next)
  {
//    ecrit_rapport("start %d, end %d\n", element->part->lba,element->part->lba+element->part->part_size-1);
    if((element->prev && (element->part->lba<=element->prev->part->lba+element->prev->part->part_size-1)) ||
	(element->next && (element->part->lba+element->part->part_size-1>=element->next->part->lba)))
    {
//      ecrit_rapport("\nChevauchement\n");
      res=1;
    }
  }
//  ecrit_rapport("Test chevauchement fin\n");
  for(element=new_list_part;element;element=new_element)
  {
    new_element=element->next;
    FREE(element);
  }
  return res;
}

t_list_part search_part(t_param_disk disk_car, const int paranoid, const int debug, const int dump_ind, const int fast_mode,const int interface)
{
  int found;
  int part_offset=PART_OFFSET128;
  int ind_stop=FALSE;
  t_CHS start;
  unsigned int old_cylinder=1234;
  t_list_part list_part=NULL;
  t_diskext partition=partition_new(disk_car->disk);
#ifdef DJGPP
  disk_car->nbr_err=0;
#endif
  start.cylinder=0;
  start.head=1;
  start.sector=1;
  partition->lba=CHS2LBA(disk_car,&start);
  if(interface)
  {
    move(22,0);
    standout();
    addstr("  Stop  ");
    standend();
  }
  aff_buffer(BUFFER_RESET,"Q");
  ecrit_rapport("\nsearch_part()\n");
  aff_disk_rapport(disk_car);
  /* ncurses: getch will be a non-blocking call */
  while((partition->lba<=disk_car->lba)&&!ind_stop)
  {
    dword old_lba=partition->lba;
    int res;
    LBA2CHS(disk_car,partition->lba,&start);
    if(interface && (old_cylinder!=start.cylinder))
    {
      move(ANALYSE_Y,ANALYSE_X);
      clrtoeol();
      doprintf("Analyse cylinder %5d/%d",start.cylinder,disk_car->CHS.cylinder);
      //    ecrit_rapport("CHS %d/%d/%d %ld %d\n",start.cylinder,start.head,start.sector,partition->lba,part_offset);
      refresh();
      old_cylinder=start.cylinder;
    }
    found=0;
    partition_reset(partition);
    switch(part_offset)
    {
      case PART_OFFSET128:
	res=search_type_128(disk_car,partition,debug,dump_ind,paranoid);
	part_offset=PART_OFFSET2;
	break;
      case PART_OFFSET2:
	res=search_type_2(disk_car,partition,debug,dump_ind,paranoid);
	part_offset=PART_OFFSET1;
	break;
      case PART_OFFSET1:
	res=search_type_1(disk_car,partition,debug,dump_ind,paranoid);
	part_offset=PART_OFFSET0;
	break;
      case PART_OFFSET0:
	res=search_type_0(disk_car,partition,debug,dump_ind,paranoid);
	part_offset=PART_QUIT;
	break;
      case PART_QUIT:
	ecrit_rapport("PART_QUIT: do nothing\n");
	res=0;
	break;
      default:
	ecrit_rapport("\nBUG: unusual value of part_offset\n");
	exit(1);
    }
    if(res<0)
    {
      if(interface)
	doprintf(m_READ_ERROR_AT, start.cylinder,start.head,start.sector);
      ecrit_rapport(m_READ_ERROR_AT, start.cylinder,start.head,start.sector);
      ecrit_rapport("lba=%ld\n",partition->lba);
    }
    else
    switch(partition->part_type)
    {
      case 0:
      case P_EXTENDED:
      case P_EXTENDX:
	break;
      default:
	{
	  dword pos_fin=partition->lba+partition->part_size-1;
	  aff_part_rapport(1,disk_car,partition);
	  if(debug && partition->info[0])
	    ecrit_rapport("\n%s\n",partition->info);
	  if((partition->part_size>1)&&(pos_fin<disk_car->lba))
	  {
	    t_diskext new_partition=partition_new(0);
	    dup_t_diskext(new_partition,partition);
	    new_partition->status=STATUS_DELETED;
	    found=1;
	    check_part(disk_car,debug,new_partition);
	    aff_part_buffer(1,disk_car,new_partition);
	    list_part=insert_new_partition(list_part,new_partition);
	    if(interface)
	      aff_buffer(BUFFER_SHOW,"Q");
	  }
	}	
	break;
    }
    if((!paranoid && found) || (part_offset==PART_QUIT))
    {
      /* Update position */
      part_offset=PART_OFFSET128;
      if(found && !paranoid)
      {
	LBA2CHS(disk_car,partition->lba+partition->part_size-1,&start);
      }
      /* Fast   head=0,1,2	sector=1		HS=0,1 1,1 2,1
       * Medium head=0,1				HS=0,1 0,7 1,1 1,7 2,1 HS
       * 	head=CHS.head, sector=CHS.sector
       * Slow   head=0..CHS.head
       * 	head=CHS.head, sector=CHS.sector
       * */
      if((start.head<disk_car->CHS.head) && ((fast_mode==2)||(start.head<2)))
      {
	if((start.sector==1)&&(fast_mode>0))
	{
	  start.sector=1+6;	/* FAT 32 backup sector */
	  part_offset=PART_OFFSET0;
	}
	else
	{
	  start.head++;
	  start.sector=1;
	}
      }
      else
	if(((start.head<disk_car->CHS.head)||(start.sector<disk_car->CHS.sector))&&(fast_mode>0))
	{	/* NTFS recovery using backup sector */
	  start.head=disk_car->CHS.head;
	  start.sector=disk_car->CHS.sector;
	  part_offset=PART_OFFSET0;
	}
	else
	{ 
	  start.cylinder++;
	  start.head=0;
	  start.sector=1;
	}
      partition->lba=CHS2LBA(disk_car,&start);
      if(interface)
      {
	switch(getch_nodelay())
	{
	  case KEY_ENTER:
#ifdef DJGPP
	  case PADENTER:
#endif
	  case '\n':
	  case '\r':
	  case 's':
	  case 'S':
	    ind_stop=1;
	    break;
	}
      }
    }
    if(partition->lba<old_lba)
    { /* Correct position */
      partition->lba=old_lba;
    }
  }
  FREE(partition);
  return list_part;
}

void init_part_order(t_list_part list_part)
{
  int nbr_log=0;
  int nbr_prim=0;
  t_list_part element;
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
    {
      case STATUS_PRIM:
      case STATUS_PRIM_BOOT:
      case STATUS_EXT:
	element->part->order=++nbr_prim;
	break;
      case STATUS_LOG:
	element->part->order=(++nbr_log)+4;
	break;
    }
  }
}

void ask_mbr_order(t_param_disk disk_car,t_list_part list_part)
{
  t_diskext table[NBR_PART_MAX];
  t_diskext table2[NBR_PART_MAX];
  int nbr_prim=0;
  int i,pos=0;
  int res;
  int car;
  int quit=0;
  t_list_part element;
  /* Initialisation */
  aff_copy(stdscr);
  move(4,0);
  aff_disk(stdscr,disk_car);
  mvaddstr(5,0,m_MBR_ORDER);
  mvaddstr(6,0,m_PART_HEADER2);
  for(element=list_part;element;element=element->next)
  {
    if((element->part->order>0) && (element->part->order<5))
      table[nbr_prim++]=element->part;
  }
  /* */
  ecrit_rapport("\nSelection des partitions primaires\n");
  for(i=0;i<nbr_prim;i++)
      aff_part_rapport(2,disk_car,table[i]);
  /* */
  do
  {
    unsigned int order;
    /* sort table into table2 */
    int part=0;
    res=0;
    for(order=1;order<=4;order++)
    {
      int nbr=0;
      for(i=0;i<nbr_prim;i++)
	if(table[i]->order==order)
	{
	  table2[part++]=table[i];
	  nbr++;
	}
      res|=(nbr>1);
    }
    if(part!=nbr_prim)
    {
      ecrit_rapport("\nBUG part %d, nbr_prim %d\n", part, nbr_prim);
    }
    /* */
    for(i=0;i<nbr_prim;i++)
    {
      move(5+2+i,0);
      clrtoeol();
      if(i==pos)
	standout();
      aff_part(stdscr,2,disk_car,table2[i]);
      if(i==pos)
	standend();
    }
    move(20,0);
    if(res)
      doprintf(m_MBR_ORDER_BAD);
    else
      doprintf(m_MBR_ORDER_GOOD);
    refresh();
    car=wgetch(stdscr);
    quit=0;
    switch(car)
    {
      case KEY_UP:
	if(--pos<0)
	  pos=nbr_prim-1;
	break;
      case KEY_DOWN:
	if(++pos>=nbr_prim)
	  pos=0;
	break;
      case KEY_PPAGE:
	pos=0;
	break;
      case KEY_NPAGE:
	pos=nbr_prim-1;
	break;
      case '1':
      case '2':
      case '3':
      case '4':
	table2[pos]->order=car-'0';
	break;
      case KEY_RIGHT:
      case ' ':
      case '+':
	if(++table2[pos]->order>4)
	  table2[pos]->order=1;
	break;
      case KEY_LEFT:
      case '-':
	if(--table2[pos]->order<1)
	  table2[pos]->order=4;
	break;
      case 'q':
      case '\r':
      case '\n':
      case KEY_ENTER:
#ifdef DJGPP
      case PADENTER:
#endif
      case 'M':
	quit=1;
	break;
    }
    refresh();
  } while(res || !quit);
}


t_list_part reduce_structure(t_list_part list_part)
{ 
  t_list_part element=list_part;
  t_list_part prev=NULL;
  while(element)
  {
    t_list_part next=element->next;
    if(element->part->status==STATUS_DELETED)
    {
      if(prev==NULL)
	list_part=next;
      else
	prev->next=next;
      if(next)
	next->prev=prev;
      FREE(element->part);
      FREE(element);
    }
    else
      prev=element;
    element=next;
  }
  return list_part;
}

t_list_part add_ext_part(t_param_disk disk_car, t_list_part list_part, const int max_ext)
{
  /* list_part need to be sorted! */
  /* All extended partitions of an P_EXTENDX  are P_EXTENDED */
  t_list_part element,deb=NULL,fin=NULL;
  int nbr_entree=0;
  t_CHS start,end;
  t_diskext new_partition;
  for(element=list_part;element;element=element->next)
  {
    if(element->part->status==STATUS_LOG)
    {
      if(deb==NULL)
      {
	deb=element;
	nbr_entree++;
      }
      fin=element;
    }
    else
      nbr_entree++;
  }
  if(!deb)
    return list_part;
  start.sector=1;
  if((nbr_entree==4)||max_ext)
  {
    start.cylinder=(deb->prev==NULL?1:LBA2cylinder(disk_car,deb->prev->part->lba+deb->prev->part->part_size-1)+1);
    start.head=0;
    end.cylinder=(fin->next==NULL?disk_car->CHS.cylinder:LBA2cylinder(disk_car,fin->next->part->lba)-1); /* 8 october 2002 */
    end.head=disk_car->CHS.head;
    end.sector=disk_car->CHS.sector;
  }
  else
  {
    start.cylinder=LBA2cylinder(disk_car,deb->part->lba);
    if(LBA2head(disk_car,deb->part->lba)==0)
      ecrit_rapport("BUG while creating a minimal extended partition\n");
    start.head=LBA2head(disk_car,deb->part->lba)-1;
    end.cylinder=LBA2cylinder(disk_car,fin->part->lba+fin->part->part_size-1);
    end.head=disk_car->CHS.head;
    end.sector=disk_car->CHS.sector;
  }
  new_partition=partition_new(deb->part->disk);
  new_partition->part_type=(end.cylinder>1023?P_EXTENDX:P_EXTENDED);
  new_partition->status=STATUS_EXT;
  new_partition->lba=CHS2LBA(disk_car,&start);
  new_partition->part_size=CHS2LBA(disk_car,&end)-new_partition->lba+1;
  return insert_new_partition(list_part, new_partition);
}

int write_mbr(t_param_disk disk_car, t_list_part list_part, int ro,int debug)
{
  t_list_part element;
  unsigned char buffer[0x200];
  unsigned char buffer_org[0x200];
  if(debug)
    ecrit_rapport("\nwrite_mbr: starting...\n");
  if(read_MBR(disk_car,&buffer_org)) return 1;
  memset((void *)&buffer,0,0x200);
  if((buffer_org[0x1FE]==0x55) && (buffer_org[0x1FF]==0xAA))
  {
    memcpy(buffer,buffer_org,TAB_PART);
    buffer[0x1FE]=0x55;
    buffer[0x1FF]=0xAA;
  } else {
    write_MBR_code_aux(buffer);
  }
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
    {
      case STATUS_PRIM:
      case STATUS_PRIM_BOOT:
      case STATUS_EXT:
	if((element->part->order>=1) && (element->part->order<=4))
	{
	  partition2entree(disk_car,0,element->part, 
	      (t_entree*)&buffer[TAB_PART+(element->part->order-1)*0x10]);
	}
	break;
    }
  }
  if(debug)
  {
    int i;
    for(i=0;i<4;i++)
      aff_entree_rapport(&buffer[0x10*i+TAB_PART]);
    diff(buffer,buffer_org);
  }
  if(!ro)
    return write_MBR(disk_car,&buffer);
  return 0;
}

int diff(unsigned char buffer[0x200], unsigned char buffer_org[0x200])
{
  int j;
  if(memcmp(buffer,buffer_org,0x200))
  {
    ecrit_rapport("\nSectors are different.\n");
    ecrit_rapport("buffer_org\n");
    for(j=0;j<4;j++)
      if(os(&buffer_org[0x10*j+TAB_PART]))
	aff_entree_rapport(&buffer_org[0x10*j+TAB_PART]);
    ecrit_rapport("buffer\n");
    for(j=0;j<4;j++)
      if(os(&buffer[0x10*j+TAB_PART]))
	aff_entree_rapport(&buffer[0x10*j+TAB_PART]);
    for(j=0;j<0x200;j++)
      if(buffer_org[j]!=buffer[j])
	ecrit_rapport("%02X %02X %02X\n", j, buffer_org[j], buffer[j]);
    ecrit_rapport("Sectors are different: end\n");
  }
  return 0;
}

int write_all_log(t_param_disk disk_car, t_list_part list_part, int ro, int debug)
{
  t_list_part element;
  int i;
  t_list_part pos_ext=NULL;
  t_list_part table[NBR_PART_MAX];
  int nbr_log=0;
  dword current_pos;
  t_diskext bloc_nextext;
  if(debug)
	ecrit_rapport("write_all_log: starting...\n");
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
	{
	  case STATUS_EXT:          /* Pas bootable */
		if((element->part->part_type==P_EXTENDED) ||
			(element->part->part_type==P_EXTENDX))
		  pos_ext=element;
		break;
	  case STATUS_LOG:
		table[nbr_log++]=element;
		break;
	}
  }
  if(pos_ext==NULL)
  {
    doprintf(m_NO_EXT_PART);
    return 1;
  }
  current_pos=pos_ext->part->lba;
  bloc_nextext=(t_diskext)MALLOC(sizeof(*bloc_nextext));
  bloc_nextext->part_type=P_EXTENDED;       /* Never P_EXTENDX */
  bloc_nextext->disk=disk_car->disk;
  for(i=0;i<nbr_log;i++)
  {
    /* tabext: partition etendu conteneur */
    /* tab[part[table[i]]: partition */
    int j;
    unsigned char buffer[0x200];
    unsigned char buffer_org[0x200];
    if(debug)
    {
      ecrit_rapport("write_all_log: CHS: %u %u %u,lba=%lu\n", LBA2cylinder(disk_car,current_pos), LBA2head(disk_car,current_pos), LBA2sector(disk_car,current_pos),current_pos);
    }
    if(hd_read_LBA(disk_car,1, &buffer_org, current_pos))
      return 1;
    memset(buffer,0,0x200);
    memcpy(buffer,buffer_org,TAB_PART);
    buffer[0x1FE]=0x55;
    buffer[0x1FF]=0xAA;
    partition2entree(disk_car,current_pos,table[i]->part, (t_entree*)&buffer[TAB_PART+0x0]);
    if(i<nbr_log-1)
    { /* Construit le pointeur vers la prochaine partition logique */
      t_CHS nextext_start;
      LBA2CHS(disk_car,table[i+1]->part->lba,&nextext_start);
      nextext_start.head=0;
      nextext_start.sector=1;
      bloc_nextext->lba=CHS2LBA(disk_car,&nextext_start);
      bloc_nextext->part_size=(table[i+1]->part->lba+table[i+1]->part->part_size-1)-bloc_nextext->lba+1;
      partition2entree(disk_car,pos_ext->part->lba,bloc_nextext, (t_entree*)&buffer[TAB_PART+0x10]);
    }
    if(ro)
    {
      for(j=0;j<4;j++)
	if(os(&buffer[0x10*j+TAB_PART]))
	  aff_entree_rapport(&buffer[0x10*j+TAB_PART]);
      diff(buffer,buffer_org);
    }
    else
    {
      if(hd_write_LBA(disk_car,1, &buffer, current_pos))
	doprintf(m_PART_WR_ERR);
    }
    current_pos=bloc_nextext->lba;
  }
  FREE(bloc_nextext);
  return 0;
}

t_list_part do_recover_cih(t_param_disk disk_car,t_list_part list_part,int debug,const int dump_ind)
{ /* tab_part need to be sorted */
  if((list_part==NULL) || (LBA2cylinder(disk_car,list_part->part->lba)>0))
  {
    t_CHS start,end;
    t_diskext partition;
    start.cylinder=0;
    start.head=1;
    start.sector=1;
    dup_t_CHS(&end,&disk_car->CHS);
    if(list_part)
    {
      end.cylinder=LBA2cylinder(disk_car,list_part->part->lba)-1;
    }
    partition=partition_new(disk_car->disk);
    partition->lba=CHS2LBA(disk_car,&start);
    partition->part_size=CHS2LBA(disk_car,&end)-partition->lba+1;
    if(rebuild_FAT_BS(disk_car,partition,debug,dump_ind,1)==0)
    {
      ecrit_rapport("rebuild_FAT_BS Ok\n");
      return insert_new_partition(list_part,partition);
    }
    else
      FREE(partition);
  }
  {
    t_diskext partition;
    t_list_part element,last=NULL;
    for(element=list_part;element;element=element->next)
      last=element;
    if(last)
    {
      partition=partition_new(disk_car->disk);
      partition->lba=last->part->lba+last->part->part_size-1;
      partition->part_size=disk_car->lba-partition->lba+1;
      if(rebuild_FAT_BS(disk_car,partition,debug,dump_ind,1)==0)
      {
	ecrit_rapport("rebuild_FAT_BS Ok\n");
	return insert_new_partition(list_part,partition);
      }
      else
	FREE(partition);
    }
  }
  return list_part;
}


int interface_recovery(t_param_disk disk_car,int paranoid,const int debug,const int dump_ind, int fast_mode,int align, const int recover_cih,const int ask_part_order, const int max_ext)
{
  int res_interface_write;
  do
  {
    t_list_part list_part;
    t_list_part element;
    int can_search_deeper=1;
    aff_copy(stdscr);
    move(4,0);
    aff_disk(stdscr,disk_car);
    move(5,0);
    res_interface_write=0;
    if(paranoid==1 && fast_mode==2)
      can_search_deeper=0;
    list_part=search_part(disk_car,paranoid,debug,dump_ind,fast_mode,1);
    if(align)
      align_structure(disk_car,list_part);
    /* Little trick to detect Linux software Raid 1 quickly */
    for(element=list_part;element;element=element->next)
    {
      unsigned char buffer[8*0x200];
      struct mdp_superblock_s *sb=(struct mdp_superblock_s*)&buffer;
      dword offset=MD_NEW_SIZE_SECTORS(element->part->part_size);
      
      if(!hd_read_LBA(disk_car,8, &buffer, element->part->lba+offset))
      {
	t_diskext new_partition=partition_new(disk_car->disk);
	dup_t_diskext(new_partition,element->part);
	if(test_MD(disk_car,sb,new_partition,debug,dump_ind)==0)
	{
	  set_MD_info(disk_car,sb,new_partition,debug,dump_ind);
	  new_partition->part_type=P_RAID;
	  new_partition->part_size=(sb->size<<1)+MD_RESERVED_SECTORS;
	  list_part=insert_new_partition(list_part,new_partition);
	}
	else
	  FREE(new_partition);
      }
    }
    if(align)	/* Need to align new partition */
      align_structure(disk_car,list_part);
    init_structure(disk_car,list_part);
#ifdef DEBUG
    check_list_part(list_part);
#endif
    if(debug)
    {
      /* Ajoute dans le fichier de log les partitions trouvees */
      ecrit_rapport("\nResults");
      for(element=list_part;element;element=element->next)
	aff_part_rapport(1,disk_car,element->part);
    }
    doprintf(m_GOD_MODE);
    list_part=ask_structure(disk_car,list_part,debug);
    if(!test_structure(list_part))
    {
      list_part=reduce_structure(list_part);
#ifdef DEBUG
      check_list_part(list_part);
#endif
      /* Sort list_part */
      list_part=sort_list_part(list_part);
#ifdef DEBUG
      check_list_part(list_part);
#endif
      if(recover_cih)
      {
	list_part=do_recover_cih(disk_car,list_part,debug,dump_ind);
	list_part=sort_list_part(list_part);
      }
      /* Creer la partition etendue globale */
      list_part=add_ext_part(disk_car,list_part,max_ext);
      ecrit_rapport("\nResults with extended");
      for(element=list_part;element;element=element->next)
	aff_part_rapport(1,disk_car,element->part);
      init_part_order(list_part);
      if(ask_part_order)
      {
	/* Demande l'ordre des entrees dans le MBR */
	ask_mbr_order(disk_car,list_part);
	/* Demande l'ordre des partitions etendues */
      }
      res_interface_write=interface_write(disk_car,list_part,can_search_deeper);
      switch(res_interface_write)
      {
	case 1:
	  if(ask_confirmation("Confirm ? (Y/N)"))
	  {
	    ecrit_rapport("write!\n");
	    write_mbr(disk_car,list_part,RW,debug);
	    write_all_log(disk_car,list_part,RW,debug);
	  }
	  else
	    ecrit_rapport("Don't write, no confirmation\n");
	break;
	case 0:
	ecrit_rapport("simulate write!\n");
	write_mbr(disk_car,list_part,RO,debug);
	write_all_log(disk_car,list_part,RO,debug);
	break;
	case 2:
	paranoid=1;
	if(fast_mode<2)
	  fast_mode++;
	break;
      }
    }
    delete_list_part(list_part);
  } while(res_interface_write==2);
  return 0;
}
