/*

    File: hdaccess.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define LINUX_WRITE 1
#ifdef LINUX_WRITE
#define open_mode O_RDWR
#else
#define open_mode O_RDONLY
#endif
#include <stdio.h>
#ifdef DJGPP
#include <stdlib.h>     /* atexit */
#include <go32.h>       /* dosmemget/put */
#include <dpmi.h>
#include <bios.h>       /* bios_k* */
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "hdaccess.h"
#include "intrface.h"
#define HD_RW_BUF_SIZ 0x10
#define HDPARM_BUF_SIZ 0x1A
#define MAX_IO_NBR 3
#define MAX_HD_ERR 100
#define PARA_CMD         8
int hd_super(int cmd, t_param_disk disk_car, dword hd_offset, int nsects, void *buffer);
int hd_identify_bios(t_param_disk param_disk,const int debug,const int allow_partial_last_cylinder);
int hd_identify_enh_bios(t_param_disk param_disk,const int debug, const int allow_partial_last_cylinder);
int check_enh_bios(t_param_disk param_disk, const int debug);
static int dos_segment = 0;
static int dos_selector = 0;

void free_dos_buffer(void)
{
  __dpmi_free_dos_memory(dos_selector);
  dos_segment = dos_selector = 0;
}

int alloc_dos_buffer(void)
{
  if (dos_segment)
    return 0;
  if ((dos_segment = __dpmi_allocate_dos_memory(18*512/16, &dos_selector))== -1)
  {
    dos_segment = 0;
    return 1;
  }
  atexit(free_dos_buffer);
  return 0;
}

int hd_super(int cmd, t_param_disk disk_car, dword hd_offset, int nsects, void *buffer)
{ 
  __dpmi_regs r;
  unsigned char buf[HD_RW_BUF_SIZ];
  int dos_segment2, dos_selector2, xfer2=nsects*512;
  if((disk_car->mode==0)||(cmd==0))
  { // Limite CHS = 1023,255,63 = 8,064Mo ~= 7.8 Go
    int head, track, sector;
    sector=(hd_offset%disk_car->CHSR.sector)+1;
    hd_offset/=disk_car->CHSR.sector;
    head=hd_offset%(disk_car->CHSR.head+1);
    track=hd_offset/(disk_car->CHSR.head+1);
    if(track<1024)
      return biosdisk(cmd, disk_car->disk, head, track, sector, nsects, buffer);
    return 1;
  }
  if(dos_segment==0)
    if(alloc_dos_buffer())
      return 1;
  if ( (dos_segment2=__dpmi_allocate_dos_memory((xfer2 + 15) >> 4, &dos_selector2)) == -1 )
    return 1;
  *(word*)&buf[0]=HD_RW_BUF_SIZ;
  *(word*)&buf[2]=nsects;
  *(dword*)&buf[0x4]=dos_segment2<<16;
  *(dword*)&buf[0x8]=hd_offset;
  *(dword*)&buf[0xC]=0;

  r.x.ds = dos_segment;
  r.x.si = 0;
  r.h.dl = disk_car->disk;
  switch(cmd)
  {
    case 2:
      r.h.ah = 0x42;
      break;
    case 3:
      r.x.ax = 0x4300;
      dosmemput(buffer,xfer2,dos_segment2<<4);
      break;
	case 0x0C:
	  r.h.ah = 0x47;
	  break;
  }
  dosmemput(&buf, HD_RW_BUF_SIZ, dos_segment<<4);
  __dpmi_int(0x13, &r);
  if(cmd==2)
    dosmemget(dos_segment2<<4, xfer2, buffer);
  __dpmi_free_dos_memory(dos_selector2);
#ifdef DEBUG
  if(r.h.ah)
  {
    doprintf(" %ld err %02X %04X\n",*(dword*)&buf[0x8], r.h.ah,r.x.flags);
  }
#endif
  return r.h.ah;
}

int check_enh_bios(t_param_disk param_disk, const int debug)
{
  __dpmi_regs r;
  r.h.ah = 0x41;
  r.x.bx = 0x55AA;
  r.h.dl = param_disk->disk;
  __dpmi_int(0x13, &r);
  if(r.x.bx != 0xAA55)  /* INT 13 Extensions not installed */
  {
    if(debug)
      ecrit_rapport("Disk %02X - INT 13 Extensions not installed\n",param_disk->disk);
    return 0;
  }
  if(debug)
  {
    ecrit_rapport("Disk %02X ",param_disk->disk);
    switch(r.h.ah)
	{
	  case 0x01:
		ecrit_rapport("Enhanced BIOS 1.x");
		break;
	  case 0x20:
		ecrit_rapport("Enhanced BIOS 2.0 / EDD-1.0");
		break;
	  case 0x21:
		ecrit_rapport("Enhanced BIOS 2.1 / EDD-1.1");
		break;
	  case 0x30:
		ecrit_rapport("Enhanced BIOS EDD-3.0");
		break;
	  default:
		ecrit_rapport("Enhanced BIOS unknown %02X",r.h.ah);
		break;
	}
    if((r.x.cx & 1)!=0)
      ecrit_rapport(" - R/W/I");
    if((r.x.cx & 4)!=0)
      ecrit_rapport(" - Identify");
    ecrit_rapport("\n");
  }
  return ((r.x.cx&1)!=0);
}

int hd_identify_enh_bios(t_param_disk param_disk,const int debug,const int allow_partial_last_cylinder)
{
  int compute_LBA=0;
  int add_cylinder;
  __dpmi_regs r;
  unsigned char buf[0x200];	/* Don't change it! */
  buf[0]=HDPARM_BUF_SIZ;
  buf[1]=0;
  if(dos_segment==0)
    if(alloc_dos_buffer())
      return 1;
  r.h.ah = 0x48;
  r.x.ds = dos_segment;
  r.x.si = 0;
  r.h.dl = param_disk->disk;
  dosmemput(&buf, HDPARM_BUF_SIZ, dos_segment<<4);
  __dpmi_int(0x13, &r);
  dosmemget(dos_segment<<4, HDPARM_BUF_SIZ, &buf);
  if(r.h.ah)
    return 1;
  param_disk->CHS.cylinder=*(word*)&buf[0x04];
  param_disk->CHS.head=*(word*)&buf[0x08];
  param_disk->CHS.sector=*(word*)&buf[0x0C];
  param_disk->lba=(*(dword*)&buf[0x10]);
  if(param_disk->lba==0)
  {
    param_disk->CHS.cylinder--;
    param_disk->CHS.head--;
    compute_LBA=1;
    param_disk->lba=(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector;
    if(debug)
      ecrit_rapport("Computes LBA from CHS\n");
  }
  else
  {
    if(param_disk->CHS.cylinder && param_disk->CHS.head && param_disk->CHS.sector)
    {
      param_disk->CHS.cylinder--;
      param_disk->CHS.head--;
      /* Some bios are buggy */
      if(param_disk->lba>(param_disk->CHS.cylinder+2)*(param_disk->CHS.head+1)*param_disk->CHS.sector)
      {
	param_disk->CHS.cylinder=(param_disk->lba/(param_disk->CHS.head+1))/param_disk->CHS.sector-1;
	if(debug)
	  ecrit_rapport("Computes C from number of sectors\n");
      }
    }
    else
    {
      if(debug)
	ecrit_rapport("Computes CHS from number of sectors\n");
      param_disk->CHS.head=255-1;
      param_disk->CHS.sector=63;
      param_disk->CHS.cylinder=(param_disk->lba/(param_disk->CHS.head+1))/param_disk->CHS.sector-1;
    }
  }
  if(debug)
    ecrit_rapport("LBA %lu, computed %u (CHS=%d,%d,%d)\n",param_disk->lba, (param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector,param_disk->CHS.cylinder,param_disk->CHS.head,param_disk->CHS.sector);
  if(compute_LBA || allow_partial_last_cylinder ||
	  (param_disk->lba <= (param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector))
  {
    t_CHS pos_CHS;
    add_cylinder=0;
    dup_t_CHS(&pos_CHS,&param_disk->CHS);
    pos_CHS.cylinder++;
    if(allow_partial_last_cylinder) {
      pos_CHS.head=0;
      pos_CHS.sector=1;
    }
    param_disk->CHSR.cylinder=0;
    if(hd_read_LBA(param_disk,1,&buf, CHS2LBA(param_disk,&pos_CHS))==0)
    {
      add_cylinder=1;
      param_disk->CHS.cylinder++;
      if(debug)
	ecrit_rapport("Disk %02X hd_identify_enh_bios, BIOS hides the last cylinder.\n",param_disk->disk);
    }
  }
  if(compute_LBA)
    param_disk->lba=(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector;
  else
  {
    if(param_disk->lba < (param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector)
    {
      ecrit_rapport("Computes LBA from CHS, previous value may be false.\n");
      param_disk->lba=(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector;
    }
  }
  param_disk->CHSR.cylinder=param_disk->CHS.cylinder;
  param_disk->CHSR.head=param_disk->CHS.head;
  param_disk->CHSR.sector=param_disk->CHS.sector;
  if(debug)
  {
    aff_disk_rapport(param_disk);
    ecrit_rapport("LBA size=%lu\n",param_disk->lba);
  }
  return 0;
}

int hd_identify_bios(t_param_disk param_disk,const int debug,const int allow_partial_last_cylinder)
{
  int add_cylinder;
  unsigned char buf[0x200];
  memset(buf,0,sizeof(buf));
  if(biosdisk(PARA_CMD,param_disk->disk,0,0,1,1,buf))
    return 1;
  if(debug)
    ecrit_rapport("Disk %02X %d max %02X\n",param_disk->disk,buf[2],(0x80+buf[2]-1));
  if(param_disk->disk>(0x80+buf[2]-1))
	return 1;
  param_disk->CHS.cylinder=((buf[0] & 0x0C0)<<2)|buf[1];
  param_disk->CHS.head=buf[3];
  param_disk->CHS.sector=buf[0] & 0x3F;
  if(param_disk->CHS.head==255)
  { // Problem found by G Rowe
    ecrit_rapport("BIOS reports an invalid heads number\n");
    param_disk->CHS.head=254;
  }
  {
    add_cylinder=0;
    if(param_disk->CHS.cylinder<1023)
    {
      int res;
      if(allow_partial_last_cylinder) {
	res=biosdisk(0x0C, param_disk->disk, 0, param_disk->CHS.cylinder+1, 0, 1, NULL);
      } else {
	res=biosdisk(0x0C, param_disk->disk, 0, param_disk->CHS.cylinder+1, param_disk->CHS.head, param_disk->CHS.sector, NULL);
      }
      if(!res)
      {
	add_cylinder=1;
	param_disk->CHS.cylinder++;
	if(debug)
	  ecrit_rapport("Disk %02X hd_identify_bios, BIOS hides the last cylinder.\n",param_disk->disk);
      }
    }
  } // while(add_cylinder);
  param_disk->lba=(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector;
  param_disk->CHSR.cylinder=param_disk->CHS.cylinder;
  param_disk->CHSR.head=param_disk->CHS.head;
  param_disk->CHSR.sector=param_disk->CHS.sector;
  if(debug)
  {
    aff_disk_rapport(param_disk);
  }
  return 0;
}

int hd_read_LBA(t_param_disk disk_car,const unsigned int nbr_secteur, void *nom_buffer, dword hd_offset)
{
  if((disk_car->nbr_err>MAX_HD_ERR)&&disk_car->halt_on_errors)
  {
    if(disk_car->nbr_err==(MAX_HD_ERR+1))
    {
      ecrit_rapport("\ndisk_car->nbr_err>MAX_HD_ERR !!!\n");
      disk_car->nbr_err++;
    }
    return 1;
  }
  if((disk_car->CHSR.cylinder==0) || (hd_offset+nbr_secteur-1<=disk_car->lba))
  {
    int i,rc=4; // sector not found/read error
    //    for(i=0;(rc==4) && (i<MAX_IO_NBR);i++)
    for(i=0;rc && (rc!=1) && (i<MAX_IO_NBR);i++)
    {
      rc=hd_super(2, disk_car, hd_offset, nbr_secteur, nom_buffer);
      if(rc)
      {
	hd_super(0, disk_car, hd_offset, nbr_secteur, nom_buffer);
      }
    }
    if(rc)
    {
      disk_car->nbr_err++;
      ecrit_rapport("hd_read_LBA lba=%lu mode=%d nbr_secteur=%u, rc=%d\n",hd_offset,disk_car->mode,nbr_secteur,rc);
      ecrit_rapport("hd_read_LBA %u/%u/%u\n", LBA2cylinder(disk_car,hd_offset),LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
    }
    return rc;
  }
  return 1;
}

int hd_write_LBA(t_param_disk disk_car,const unsigned int nbr_secteur, void *nom_buffer, dword hd_offset)
{
  if((disk_car->nbr_err>MAX_HD_ERR)&&disk_car->halt_on_errors)
  {
    if(disk_car->nbr_err==(MAX_HD_ERR+1))
    {
      ecrit_rapport("\ndisk_car->nbr_err>MAX_HD_ERR !!!\n");
      disk_car->nbr_err++;
    }
    return 1;
  }
  {
    int i,rc=4; // sector not found/read error
    for(i=0;(rc==4)&&(i<MAX_IO_NBR);i++)
    {
      rc=hd_super(3, disk_car, hd_offset, nbr_secteur, nom_buffer);
      if(rc)
		ecrit_rapport("hd_write_LBA mode=%d nbr_secteur=%u, rc=%d\n",disk_car->mode,nbr_secteur,rc);
    }
    if(rc)
      disk_car->nbr_err++;
    return rc;
  }
}

int hd_identify(t_param_disk param_disk, const int debug,const int allow_partial_last_cylinder)
{
  /* standard BIOS access */
  param_disk->mode=0;
  param_disk->nbr_err=0;
  if(!hd_identify_bios(param_disk,debug,allow_partial_last_cylinder))
  {
    /* enhanced BIOS access */
    t_param_disk param_disk_enh=(t_param_disk)MALLOC(sizeof(*param_disk_enh));
    dup_t_param_disk(param_disk_enh,param_disk);
    param_disk_enh->mode=1;
    if(check_enh_bios(param_disk_enh,debug))
    {
      if(!hd_identify_enh_bios(param_disk_enh,debug,allow_partial_last_cylinder))
      {
	if((param_disk_enh->CHS.head==param_disk->CHS.head)
	    &&(param_disk_enh->CHS.sector==param_disk->CHS.sector))
	{
	  /* Remove unnecessary standard disk access */
	  dup_t_param_disk(param_disk,param_disk_enh);
	}
	else
	{
	  /* Mixed mode */
	  param_disk->mode=2; /* extended functions */
	  /* standard geometry H,S, compute C from LBA */
	  param_disk->lba=param_disk_enh->lba;
	  param_disk->CHS.cylinder=(param_disk->lba/(param_disk->CHS.head+1))/param_disk->CHS.sector-1;
	}
      }
    }
    FREE(param_disk_enh);
    return 0;
  }
  return 1;
}

int hd_parse(t_param_disk param_disk[NBR_DISK_MAX],const int debug,const int allow_partial_last_cylinder)
{
  int i;
  int nbr_disk=0;
  int ind_stop=0;
  for(i=0x80;(i<0x88)&&!ind_stop;i++)
  {
    param_disk[nbr_disk]->disk=i;
    if(!hd_identify(param_disk[nbr_disk],debug,allow_partial_last_cylinder))
      nbr_disk++;
	else
	  ind_stop=1;
  }
  return nbr_disk;
}
#elif defined(LINUX)
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/hdreg.h>
#include <fcntl.h> 
#include <sys/mount.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "hdaccess.h"

dword lseek64(int,loff_t, int);
/* if lseek don't work well, try lseek64 */

int hd_read_LBA(t_param_disk disk_car,const unsigned int nbr_secteur, void *nom_buffer, dword pos)
{
  int i;
//  ecrit_rapport("hd_read_LBA(%d,%d,buffer,%lx)\n", disk_car->disk,nbr_secteur,pos);
/*
  if(pos<1000)
  {
    memset(nom_buffer,0,nbr_secteur*0x200);
    return 0;
  }
 */
  i=lseek64(disk_car->disk,((loff_t)pos<<9)+(loff_t)disk_car->offset,SEEK_SET);
  if(i==-1)
  {
    ecrit_rapport("hd_read_LBA(%d,%d,buffer,%lu) seek err\n", disk_car->disk,nbr_secteur,pos);
    return -1;
  }
  if(read(disk_car->disk, nom_buffer, nbr_secteur*0x200)==-1)
  {
    ecrit_rapport("hd_read_LBA(%d,%d,buffer,%lu) read err\n", disk_car->disk,nbr_secteur,pos);
	return -1;
  }
  return 0;
}

int hd_write_LBA(t_param_disk disk_car,const unsigned int nbr_secteur, void *nom_buffer, dword pos)
{
#ifdef LINUX_WRITE
  int i;
  i=lseek64(disk_car->disk,((loff_t)pos<<9)+(loff_t)(disk_car->offset),SEEK_SET);
  if(i==-1)
    return -1;
  if(write(disk_car->disk, nom_buffer, nbr_secteur*0x200)<0)
    return -1;
  return 0;
#else
  return 1;
#endif
}

void test_disk_availability(int * nbr_disk,t_param_disk disk_car, const char *device,const int allow_partial_last_cylinder, const int debug);
void test_disk_availability(int * nbr_disk,t_param_disk disk_car, const char *device,const int allow_partial_last_cylinder, const int debug)
{
  int hd_h = open(device, open_mode);
  if(hd_h>0)
  {
    int ind=0;
    struct hd_geometry geometry;
#ifdef HDIO_GETGEO_BIG
    struct hd_big_geometry geometry_big;
#endif
    /* Little trick from Linux fdisk */
    /* Blocks are visible in more than one way:
       e.g. as block on /dev/hda and as block on /dev/hda3
       By a bug in the Linux buffer cache, we will see the old
       contents of /dev/hda when the change was made to /dev/hda3.
       In order to avoid this, discard all blocks on /dev/hda. */
    ioctl(hd_h, BLKFLSBUF);	/* ignore errors */
    				/* e.g. Permission Denied */
#ifdef HDIO_GETGEO_BIG
    if (!ioctl(hd_h, HDIO_GETGEO_BIG, &geometry_big))
    { /* I can get the geometry */
      if(debug)
	ecrit_rapport("test_disk_availability %s HDIO_GETGEO_BIG Ok\n",device);
      disk_car->lba=geometry_big.cylinders*geometry_big.heads*geometry_big.sectors;
      disk_car->CHS.cylinder= geometry_big.cylinders-1;
      disk_car->CHS.head=geometry_big.heads-1;
      disk_car->CHS.sector= geometry_big.sectors;
      ind=1;
    }
#endif
    if (!ind && !ioctl(hd_h, HDIO_GETGEO, &geometry))
    { /* I can get the geometry */
      if(debug)
	ecrit_rapport("test_disk_availability %s HDIO_GETGEO Ok\n",device);
      disk_car->lba=geometry.cylinders*geometry.heads*geometry.sectors;
      disk_car->CHS.cylinder= geometry.cylinders-1;
      disk_car->CHS.head=geometry.heads-1;
      disk_car->CHS.sector= geometry.sectors;
      ind=1;
    }
    if(ind)
    {
      if(allow_partial_last_cylinder)	/* A bit simple, can do better... */
	disk_car->CHS.cylinder++;
      disk_car->name=MALLOC(strlen(device)+1);
      strcpy(disk_car->name,device);
      disk_car->disk= hd_h;
      disk_car->offset=0;
      (*nbr_disk)++;
    }
    else
      close(hd_h);
  }
}

void test_file_availability(int * nbr_disk,t_param_disk disk_car, const char *device,const int allow_partial_last_cylinder, const int debug);
void test_file_availability(int * nbr_disk,t_param_disk disk_car, const char *device,const int allow_partial_last_cylinder, const int debug)
{
  int hd_h = open(device, open_mode);
  if(hd_h>0)
  {
    unsigned char buffer[0x200];
    if(read(hd_h,buffer,sizeof(buffer))==sizeof(buffer))
    {
      if(strncmp(buffer,"DOSEMU",6)==0){
	disk_car->offset=*(u_long*)(buffer+19);
	disk_car->CHS.head=*(u_long*)(buffer+7)-1;
	disk_car->CHS.sector=*(u_long*)(buffer+11);
	disk_car->CHS.cylinder=*(u_long*)(buffer+15)-1;
	disk_car->lba=(disk_car->CHS.cylinder+1)*(disk_car->CHS.head+1)*disk_car->CHS.sector;
      }
      else
      {
	struct stat stat_rec;
	fstat(hd_h,&stat_rec);
      	disk_car->offset=0;
	disk_car->CHS.head=255-1;
	disk_car->CHS.sector=63;
	disk_car->lba=stat_rec.st_size/512;
	disk_car->CHS.cylinder=disk_car->lba/disk_car->CHS.sector/(disk_car->CHS.head+1)-1+1;	// Add an extra cylinder because file is often truncated.
      }
      disk_car->name=MALLOC(strlen(device)+1);
      strcpy(disk_car->name,device);
      disk_car->disk= hd_h;
      (*nbr_disk)++;
    }
    else
      close(hd_h);
  }
}

int hd_parse(t_param_disk param_disk[NBR_DISK_MAX],const int debug,const int allow_partial_last_cylinder)
{
  int i;
  char device_ide[]="/dev/hda";
  char device_scsi[]="/dev/sda";
  char device_ida[]="/dev/ida/c0d0";
  char device_rd[]="/dev/rd/c0d0";
  int nbr_disk=0;
  /* Disk IDE */
  for(i=0;i<8;i++)
  {
    device_ide[strlen(device_ide)-1]='a'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ide,allow_partial_last_cylinder,debug);
  }
  /* Disk SCSI */
  for(i=0;i<8;i++)
  {
    device_scsi[strlen(device_scsi)-1]='a'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_scsi,allow_partial_last_cylinder,debug);
  }
  /* Device RAID Compaq */
  for(i=0;i<8;i++)
  {
    device_ida[strlen(device_ida)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ida,allow_partial_last_cylinder,debug);
  }
  /* Device RAID */
  for(i=0;i<8;i++)
  {
    device_rd[strlen(device_rd)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_rd,allow_partial_last_cylinder,debug);
  }
  test_file_availability(&nbr_disk,param_disk[nbr_disk], "hdimage",allow_partial_last_cylinder,debug);
  return nbr_disk;
}

int hd_identify(t_param_disk param_disk, const int debug,const int allow_partial_last_cylinder)
{
  char device_ide[]="/dev/hda";
  int nbr_disk=0;
  device_ide[strlen(device_ide)-1]='a'+(param_disk->disk-0x80);
  test_disk_availability(&nbr_disk,param_disk, device_ide,allow_partial_last_cylinder,debug);
  return !nbr_disk;
}
#else
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h> 
#include <sys/disklabel.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "hdaccess.h"
int hd_read_LBA(t_param_disk disk_car,const unsigned int nbr_secteur, void *nom_buffer, dword pos)
{
  int i;
  i=lseek(disk_car->disk,(loff_t)(pos)<<9,SEEK_SET);
  if(i==-1)
  {
    ecrit_rapport("hd_read_LBA(%u,%u,buffer,%lu) seek err\n", disk_car->disk,nbr_secteur,pos);
    return -1;
  }
  if(read(disk_car->disk, nom_buffer, nbr_secteur*0x200)==-1)
  {
    ecrit_rapport("hd_read_LBA(%u,%u,buffer,%lu) read err\n", disk_car->disk,nbr_secteur,pos);
	return -1;
  }
  return 0;
}

int hd_write_LBA(t_param_disk disk_car,const unsigned int nbr_secteur, void *nom_buffer, dword pos)
{
#ifdef LINUX_WRITE
  int i;
  i=lseek(disk_car->disk,pos<<9,SEEK_SET);
  if(i==-1)
    return -1;
  if(write(disk_car->disk, nom_buffer, nbr_secteur*0x200)<0)
    return -1;
  return 0;
#else
  return 1;
#endif
}

void test_disk_availability(int * nbr_disk,t_param_disk disk_car, const char *device,const int allow_partial_last_cylinder, const int debug);
void test_disk_availability(int * nbr_disk,t_param_disk disk_car, const char *device,const int allow_partial_last_cylinder, const int debug)
{
  struct disklabel geometry;
  int hd_h = open(device, open_mode);
  if(hd_h>0)
  {
    if (!ioctl(hd_h, DIOCGDINFO, &geometry))
    { /* I can get the geometry */
      disk_car->name=MALLOC(strlen(device)+1);
      strcpy(disk_car->name,device);
/*
      disk_car->CHS.cylinder= geometry.cylinders-1;
      disk_car->CHS.head=geometry.heads-1;
      disk_car->CHS.sector= geometry.sectors;
 */
      disk_car->lba=geometry.d_ncylinders*geometry.d_ntracks*geometry.d_nsectors;
      disk_car->CHS.cylinder=geometry.d_ncylinders-1;
      disk_car->CHS.head=geometry.d_ntracks-1;
      disk_car->CHS.sector=geometry.d_nsectors;
      disk_car->disk= hd_h;
      if(allow_partial_last_cylinder)	/* A bit simple, can do better... */
	disk_car->CHS.cylinder++;
      (*nbr_disk)++;
    }
    else
      close(hd_h);
  }
}

int hd_parse(t_param_disk param_disk[NBR_DISK_MAX],const int debug,const int allow_partial_last_cylinder)
{
  /* Need to check http://mattriffle.com/mirrors/freebsd/doc/en_US.ISO8859-1/books/handbook/disks-naming.html#DISK-NAMING-PHYSICAL-TABLE */
  int i;
  char device_ide[]= "/dev/rwd0";	/* raw winchester disk */
  char device_ide2[]="/dev/rad0";
  char device_scsi[]="/dev/rda0";	/* raw scsci disk */
  char device_optdisk[]="/dev/rod0";
/* wd da */
  int nbr_disk=0;
  /* Disk IDE */
  for(i=0;i<8;i++)
  {
    device_ide[strlen(device_ide)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ide,allow_partial_last_cylinder,debug);
  }
  for(i=0;i<8;i++)
  {
    device_ide2[strlen(device_ide2)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ide2,allow_partial_last_cylinder,debug);
  }
  /* Disk SCSI */
  for(i=0;i<8;i++)
  {
    device_scsi[strlen(device_scsi)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_scsi,allow_partial_last_cylinder,debug);
  }
  /* optical disks */
  for(i=0;i<8;i++)
  {
    device_optdisk[strlen(device_scsi)-1]='a'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_optdisk,allow_partial_last_cylinder,debug);
  } 
  return nbr_disk;
}
#endif

