/*

    File: intrface.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "types.h"
#include "common.h"
#include "lang.h"
#include "intrface.h"
#include "readpart.h"
#include "godmode.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "lang.h"
#include "hdaccess.h"
#include "fat.h"
#include "ext2.h"
#include "rfs.h"
#include "ntfs.h"
#include "adv.h"

extern char* nom_os[256];
int get_string(char *str, int len, char *def);
void interface_options(int *paranoid,int *dump_ind,int *fast_mode,int *align,int *recover_cih,int *allow_partial_last_cylinder,int *ask_part_order,int *halt_on_errors, int *max_ext);
void interface_editor_position(const t_param_disk param_disk, unsigned long int *lba);
void interface_editor(t_param_disk param_disk);
#define GS_DEFAULT -1
#define GS_ESCAPE -2
int dump_editor(const unsigned char *nom_dump,const unsigned int lng, const int menu_pos);
void aff_CHS_from_LBA(WINDOW *window, const t_param_disk disk_car, dword hd_offset);
void aff_CHS_from_LBA_buffer(const t_param_disk disk_car, dword hd_offset);
void aff_CHS_from_LBA_rapport(const t_param_disk disk_car, dword hd_offset);
int change_part_type(const t_param_disk disk_car,t_diskext partition);


int get_string(char *str, int len, char *def)
{
    int c;
    int i = 0;
    int x, y;
    int use_def = FALSE;

    getyx(stdscr, y, x);
    clrtoeol();

    str[0] = 0;

    if (def != NULL) {
	mvaddstr(y, x, def);
	move(y, x);
	use_def = TRUE;
    }

    refresh();
    while ((c = getch()) != '\n' && c != CR) {
	switch (c) {
	case ESC:
	    move(y, x);
	    clrtoeol();
	    refresh();
	    return GS_ESCAPE;
	case KEY_DC:
	case KEY_BACKSPACE:
	    if (i > 0) {
		str[--i] = 0;
		mvaddch(y, x+i, ' ');
		move(y, x+i);
	    } else if (use_def) {
		clrtoeol();
		use_def = FALSE;
	    }
	    break;
	default:
	    if (i < len && isprint(c)) {
		mvaddch(y, x+i, c);
		if (use_def) {
		    clrtoeol();
		    use_def = FALSE;
		}
		str[i++] = c;
		str[i] = 0;
	    }
	}
	refresh();
    }

    if (use_def)
	return GS_DEFAULT;
    else
	return i;
}

int getch_nodelay()
{
  int res;
  nodelay(stdscr,TRUE);
  res=getch();
  nodelay(stdscr,FALSE);
  return res;
}

/*
 * Actual function which prints the button bar and highlights the active button
 * Should not be called directly. Call function menuSelect instead.
 */

int menuUpdate( int y, int x, const struct MenuItem *menuItems, const unsigned int itemLength, const char *available, const int menuType, int current )
{
  return wmenuUpdate(stdscr,y,x,menuItems,itemLength,available,menuType,current);
}

int wmenuUpdate(WINDOW *window, int y, int x, const struct MenuItem *menuItems, const unsigned int itemLength, const char *available, const int menuType, int current)
{
    int i, lmargin = x, ymargin = y;
    const char *mcd;

    /* Print available buttons */
    for( i = 0; menuItems[i].key; i++ )
    {
      char buff[80];
      unsigned int lenName;
      const char *mi;
      wmove(window, y, x ); wclrtoeol(window);

	/* Search next available button */
	while( menuItems[i].key && !strchr(available, menuItems[i].key) )
	{
	    i++;
	}
	if( !menuItems[i].key ) break; /* No more menu items */

	/* If selected item is not available and we have bypassed it,
	   make current item selected */
	if( current < i && menuItems[current].key < 0 ) current = i;

	/* If current item is selected, highlight it */
	if( current == i ) /*attron( A_REVERSE )*/ wstandout (window);

	/* Print item */
	mi = menuItems[i].name;
	lenName = strlen( mi );
	if(lenName>=sizeof(buff))
	{
	  ecrit_rapport("\nBUG: %s\n",mi);
	}
#ifdef DJGPP
	if(lenName > itemLength)
	{
	  if( menuType & MENU_BUTTON )
		sprintf(buff, "[%s]",mi);
	  else
		sprintf(buff, "%s",mi);
	}
	else
	{
	  if( menuType & MENU_BUTTON )
		sprintf( buff, "[%*s%-*s]", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	  else
		sprintf( buff, "%*s%-*s", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	}
#else
	if(lenName > itemLength)
	{
	  if( menuType & MENU_BUTTON )
		snprintf(buff, sizeof(buff),"[%s]",mi);
	  else
		snprintf(buff, sizeof(buff),"%s",mi);
	}
	else
	{
	  if( menuType & MENU_BUTTON )
		snprintf( buff, sizeof(buff),"[%*s%-*s]", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	  else
		snprintf( buff, sizeof(buff),"%*s%-*s", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	}
#endif
	mvwaddstr(window, y, x, buff );

	/* Lowlight after selected item */
	if( current == i ) /*attroff( A_REVERSE )*/ wstandend (window);

	/* Calculate position for the next item */
	if( menuType & MENU_VERT )
	{
	    y += 1;
	    if( y >= WARNING_START )
	    {
		y = ymargin;
		x += itemLength + MENU_SPACING;
		if( menuType & MENU_BUTTON ) x += 2;
	    }
	}
	else
	{
	    x += itemLength + MENU_SPACING;
	    if( menuType & MENU_BUTTON ) x += 2;
	    if( x > COLUMNS - lmargin - 12 )
	    {
		x = lmargin;
		y ++ ;
	    }
	}
    }
    /* Print the description of selected item */
    mcd = menuItems[current].desc;
    mvwaddstr(window, WARNING_START + 1, (COLUMNS - strlen( mcd )) / 2, mcd );
    return y;
}

/* This function takes a list of menu items, lets the user choose one *
 * and returns the value keyboard shortcut of the selected menu item  */

int menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
		const char *available, int menuType, int menuDefault )
{
  return wmenuSelect(stdscr,y,x,menuItems,itemLength,available,menuType,menuDefault);
}

int wmenuSelect(WINDOW *window, int y, int x, struct MenuItem *menuItems, int itemLength, const char *available, int menuType, int menuDefault)
{
    int i, ylast = y, key = 0, current = menuDefault;
    if( !( menuType & ( MENU_HORIZ | MENU_VERT ) ) )    
    {
	wdoprintf(window,"Menu without direction. Defaulting horizontal.");
	menuType |= MENU_HORIZ;
    }
    /* Make sure that the current is one of the available items */
    while( !strchr(available, menuItems[current].key) )
    {
	current ++ ;
	if( !menuItems[current].key ) current = 0;
    }
    /* Repeat until allowable choice has been made */
    while( !key )
    {
	/* Display the menu */
	ylast = wmenuUpdate( window, y, x, menuItems, itemLength, available,
			    menuType, current );
	wrefresh(window);
	/* Don't put wgetch after the following wclrtoeol */
	key = wgetch(window);

	/* Clear out all prompts and such */
	for( i = y; i < ylast; i ++ )
	{
	    wmove(window, i, x );
	    wclrtoeol(window);
	}
	wmove(window, WARNING_START + 1, 0 );
	wclrtoeol(window);
	/* Cursor keys */
	switch(key)
	{
	  case KEY_UP:
	    if( menuType & MENU_VERT )
	    {
	      do {
		current -- ;
		if( current < 0 ) while( menuItems[current+1].key ) current ++ ;
	      } while( !strchr( available, menuItems[current].key ) );
	      key = 0;
	    }
	    else
	      key = MENU_UP;
	    break;
	  case KEY_DOWN:
	    if( menuType & MENU_VERT )
	    {
	      do {
		current ++ ;
		if( !menuItems[current].key ) current = 0 ;
	      } while( !strchr( available, menuItems[current].key ) );
	      key = 0;
	    }
	    else
	      key = MENU_DOWN;
	    break;
	  case KEY_RIGHT:
	    if( menuType & MENU_HORIZ )
	    {
	      do {
		current ++ ;
		if( !menuItems[current].key ) 
		{
		  current = 0 ;
		}
	      } while( !strchr( available, menuItems[current].key ) );
	      key = 0;
	    }
	    else
	      key = MENU_RIGHT;
	    break;
	  case KEY_LEFT:
	    if( menuType & MENU_HORIZ )
	    {
	      do {
		current -- ;
		if( current < 0 )
		{
		  while( menuItems[current + 1].key ) current ++ ;
		}
	      } while( !strchr( available, menuItems[current].key ) );
	      key = 0;
	    }
	    else
	      key = MENU_LEFT;
	    break;
	}
	/* Enter equals to the keyboard shortcut of current menu item */
	if((key==13) || (key==10) || (key==KEY_ENTER) ||(((menuType & MENU_HORIZ) == 0) && (key==MENU_RIGHT || key==MENU_LEFT)))
	    key = menuItems[current].key;
#ifdef DJGPP
	if(key==PADENTER)
	  key = menuItems[current].key;
#endif

	/* Should all keys to be accepted? */
	if( key && (menuType & MENU_ACCEPT_OTHERS) ) break;
	/* Is pressed key among acceptable ones */
	if( key && (strchr(available, tolower(key)) || strchr(available, key)))
	    break;
	/* The key has not been accepted so far -> let's reject it */
#ifdef DEBUG
	if( key )
	{
		wmove(window,5,0);
		wdoprintf(window,"key %03X",key);
	    key = 0;
	    putchar( BELL );
	}
#endif
    }
    /* Clear out prompts and such */
    for( i = y; i <= ylast; i ++ )
    {
	wmove(window, i, x );
	wclrtoeol(window);
    }
    wmove(window, WARNING_START + 1, 0 );
    wclrtoeol(window);
    return key;
}

/* A function which displays "Press a key to continue"                  *
 * and waits for a keypress.                                            *
 * Perhaps calling function menuSelect is a bit overkill but who cares? */

void menuContinue(void)
{
    static struct MenuItem menuContinueBtn[]=
    {
	{ 'c', "", "Press a key to continue" },
	{ 0, NULL, NULL }
    };

    menuSelect(23, 0,
	menuContinueBtn, 0, "c", MENU_HORIZ | MENU_ACCEPT_OTHERS, 0 );
}

/* Function menuSelect takes way too many parameters  *
 * Luckily, most of time we can do with this function */

int menuSimple(struct MenuItem *menuItems, int menuDefault)
{
    int i, j, itemLength = 0;
    char available[MENU_MAX_ITEMS];

    for(i = 0; menuItems[i].key; i++)
    {
	j = strlen(menuItems[i].name);
	if( j > itemLength ) itemLength = j;
	available[i] = menuItems[i].key;
    }
    available[i] = 0;
    return menuSelect(18, 0, menuItems, itemLength, available, MENU_HORIZ | MENU_BUTTON, menuDefault);
}

/* End of command menu support code */

void change_geometry(t_param_disk param_disk)
{
  int done = FALSE;
  char def[LINE_LENGTH];
  char response[LINE_LENGTH];
  int tmp_val;
  int command;

  while (!done) {
	static struct MenuItem menuGeometry[]=
	{
	  { 'c', "Cylinders", "Change cylinder geometry" },
	  { 'h', "Heads", "Change head geometry" },
	  { 's', "Sectors", "Change sector geometry" },
	  { 'd', "Done", "Done with changing geometry" },
	  { 0, NULL, NULL }
	};
	move(COMMAND_LINE_Y, COMMAND_LINE_X);
	clrtoeol();
	refresh();
	command=menuSimple(menuGeometry, 3);
	command=toupper(command);
	switch (command) {
	  case 'C':
		sprintf(def, "%d", param_disk->CHS.cylinder+1);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of cylinders: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_CYLINDERS) {
			param_disk->CHS.cylinder = tmp_val-1;
		  } else
			doprintf("Illegal cylinders value");
		}
		break;
	  case 'H':
		sprintf(def, "%d", param_disk->CHS.head+1);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of heads: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_HEADS) {
		    param_disk->CHS.head = tmp_val-1;
		    param_disk->CHS.cylinder=(param_disk->lba/(param_disk->CHS.head+1))/param_disk->CHS.sector-1;
		  } else
			doprintf("Illegal heads value");
		}
		break;
	  case 'S':
		sprintf(def, "%d", param_disk->CHS.sector);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of sectors per track: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_SECTORS) {
			param_disk->CHS.sector = tmp_val;
		  } else
			doprintf("Illegal sectors value");
		}
		break;
	  case ESC:
	  case 'D':
		done = TRUE;
		break;
	}
  }
}

unsigned int ask_number(const unsigned int val_cur, const unsigned int val_min, const unsigned int val_max, const char * _format, ...)
{
  char res[200];
  char res2[200];
  char response[LINE_LENGTH];
  char def[LINE_LENGTH];
  unsigned int tmp_val;
  va_list ap;
  va_start(ap,_format);
#ifdef DJGPP
  vsprintf(res,_format,ap);
  sprintf(res2,"(%u-%u) :",val_min,val_max);
#else
  vsnprintf(res,sizeof(res),_format,ap);
  snprintf(res2,sizeof(res2),"(%u-%u) :",val_min,val_max);
#endif
  va_end(ap);

  mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, res);
  addstr(res2);
  sprintf(def, "%u", val_cur);
  if (get_string(response, LINE_LENGTH, def) > 0)
  {
    tmp_val = atoi(response);
    if (tmp_val >= val_min && tmp_val <= val_max)
      return tmp_val;
  }
  return val_cur;
}

t_list_part add_partition(const t_param_disk disk_car,t_list_part list_part, const int debug)
{
  t_CHS start,end;
  t_diskext new_partition=partition_new(disk_car->disk);
  int position=0;
  start.cylinder=0;
  start.head=0;
  start.sector=1;
  end.cylinder=disk_car->CHS.cylinder;
  end.head=disk_car->CHS.head;
  end.sector=disk_car->CHS.sector;
  {
    int done = FALSE;
    while (!done) {
      int command;
      static struct MenuItem menuGeometry[]=
      {
	{ 'c', "Cylinders", 	"Change starting cylinder" },
	{ 'h', "Heads", 	"Change starting head" },
	{ 's', "Sectors", 	"Change starting sector" },
	{ 'C', "Cylinders", 	"Change ending cylinder" },
	{ 'H', "Heads", 	"Change ending head" },
	{ 'S', "Sectors", 	"Change ending sector" },
	{ 'T' ,"Type",		"Change partition type"},
	{ 'd', "Done", "" },
	{ 0, NULL, NULL }
      };
      aff_copy(stdscr);
      move(4,0);
      aff_disk(stdscr,disk_car);
      new_partition->lba=CHS2LBA(disk_car,&start);
      new_partition->part_size=CHS2LBA(disk_car,&end) - new_partition->lba + 1;
      move(10, 0);
      clrtoeol();
      aff_part(stdscr,0,disk_car,new_partition);
      move(COMMAND_LINE_Y, COMMAND_LINE_X);
      clrtoeol();
      refresh();
      command=menuSimple(menuGeometry, position);
      switch (command) {
	case 'c':
	  start.cylinder=ask_number(start.cylinder,0,disk_car->CHS.cylinder,"Enter the starting cylinder ");
	  position=1;
	  break;
	case 'h':
	  start.head=ask_number(start.head,0,disk_car->CHS.head,"Enter the starting head ");
	  position=2;
	  break;
	case 's':
	  start.sector=ask_number(start.sector,1,disk_car->CHS.sector,"Enter the starting sector ");
	  position=3;
	  break;
	case 'C':
	  end.cylinder=ask_number(end.cylinder,start.cylinder,disk_car->CHS.cylinder,"Enter the ending cylinder ");
	  position=4;
	  break;
	case 'H':
	  end.head=ask_number(end.head,0,disk_car->CHS.head,"Enter the ending head ");
	  position=5;
	  break;
	case 'S':
	  end.sector=ask_number(end.sector,1,disk_car->CHS.sector,"Enter the ending sector ");
	  position=6;
	  break;
	case 'T':
	case 't':
	  change_part_type(disk_car,new_partition);
	  position=7;
	  break;
	case ESC:
	case 'd':
	case 'D':
	case 'q':
	case 'Q':
	  done = TRUE;
	  break;
      }
    }
  }
  if(CHS2LBA(disk_car,&end)>new_partition->lba)
    return insert_new_partition(list_part, new_partition);
  FREE(new_partition);
  return list_part;
}

void aff_copy_full(WINDOW *window)
{
  wclear(window);
  keypad(window, TRUE); // Need it to get arrow key
  mvwaddstr(window,0,0,m_Copyright);
#ifdef DJGPP
  wdoprintf(window,"Dos version\n\n");
#elif defined(BSD)
  wdoprintf(window,"BSD version\n\n");
#elif defined(LINUX)
  wdoprintf(window,"Linux version\n\n");
#else
  wdoprintf(window,"Undefined OS\n\n");
#endif
}

void aff_copy(WINDOW *window)
{
  wclear(window);
  keypad(window, TRUE); // Need it to get arrow key
  mvwaddstr(window,0,0,m_Copyright);
}

int do_curses_testdisk(int debug,int paranoid,int dump_ind, int fast_mode,int align,int do_analyse)
{
  int i;
  int command;
  int nbr_disk;
  int current_disk=0;
  int done=0;
  int recover_cih=0;
  int allow_partial_last_cylinder=0;
  int ask_part_order=0;
  int halt_on_errors=1;
  t_param_disk param_disk[NBR_DISK_MAX];
  int max_ext=0;
  static struct MenuItem menuMain[]=
  {
	{'a',"Analyse","Analyse current partition structure and search for lost partitions"},
	{'d',"Delete","Delete all data in the partition table"},
	{'c',"MBR Code","Write the Classic MBR code to first sector"},
	{'g',"Geometry", "Change disk geometry (expert only)" },
	{'o',"Options","Modify options"},
	{'e',"Editor","Basic disk editor"},
	{'t',"Advanced","Advanced boot recovery"},
	{'q',"Quit","Quit program"},
	{0,NULL,NULL}
  };
//  use_env(TRUE);
  initscr();
  noecho();
  keypad(stdscr, TRUE); // Need it to get arrow key
#ifndef DJGPP
  nonl(); //don't use for Dos version but enter will work with it... dilema
#endif
  crmode();
  //  intrflush(stdscr, FALSE);
  cbreak();
  if(LINES<25)
  {
    nl();
    endwin();
    return 1;
  }
  for(i=0;i<NBR_DISK_MAX;i++)
  {
	param_disk[i]=(t_param_disk)MALLOC(sizeof(*param_disk[i]));
#ifndef DJGPP
	param_disk[i]->name=NULL;
#endif
  }
  nbr_disk=hd_parse(param_disk,debug,allow_partial_last_cylinder);
  /* save disk parameters to rapport */
  ecrit_rapport("Hard disk list\n");
  for(i=0;i<nbr_disk;i++)
    aff_disk_rapport(param_disk[i]);
  /* ncurses interface */
  if(nbr_disk>0)
  {
    while(!done)
    {
      aff_copy_full(stdscr);
      move(5,0);
      for(i=0;i<nbr_disk;i++)
      {
	if(i!=current_disk)
	  aff_disk(stdscr,param_disk[i]);
	else
	{
	  standout();
	  aff_disk(stdscr,param_disk[i]);
	  standend();
	}
      }
#ifdef DJGPP
      param_disk[current_disk]->halt_on_errors=halt_on_errors;
#endif
      command = toupper(menuSelect(INTER_MAIN_Y, INTER_MAIN_X, menuMain, 8,
	  "agdcotq", MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 0));
      /* e for editor will be added when the editor will be better */
      switch(command)
      {
	case 'G':
	  change_geometry(param_disk[current_disk]);
	  break;
	case 'D':
	  move(INTER_MAIN_Y,INTER_MAIN_X);
	  write_clean_table(param_disk[current_disk]);
	  break;
	case 'C':
	  move(INTER_MAIN_Y,INTER_MAIN_X);
	  write_MBR_code(param_disk[current_disk]);
	  break;
	case 'A':
	  if(do_analyse)
	    interface_analyse(param_disk[current_disk],debug);
	  interface_recovery(param_disk[current_disk],paranoid,debug, dump_ind, fast_mode,align,recover_cih,ask_part_order,max_ext);
	  break;
	case 'T':	/* Test */
	  interface_adv(param_disk[current_disk], debug, dump_ind);
	  break;
	case 'E':
	  interface_editor(param_disk[current_disk]);
	  break;
	case 'Q':
	  done = 1;
	  break;
	case 'O':
	  {
	    int old_allow_partial_last_cylinder=allow_partial_last_cylinder;
	    interface_options(&paranoid,&dump_ind,&fast_mode,&align,&recover_cih,&allow_partial_last_cylinder,&ask_part_order,&halt_on_errors,&max_ext);
	    if(old_allow_partial_last_cylinder!=allow_partial_last_cylinder)
	      hd_parse(param_disk,debug,allow_partial_last_cylinder);
	  }
	  break;
	case MENU_UP:
	  if(current_disk>0)
	    current_disk--;
	  break;
	case MENU_DOWN:
	  if(current_disk<nbr_disk-1)
	    current_disk++;
	  break;
#ifdef DEBUG
	default:
	  move(10,0);
	  doprintf("key %03X %d", command, command);
	  refresh();
	  break;
#endif
      }
    }
  }
  nl();
  endwin();
  ecrit_rapport("\n");
  for(i=0;i<NBR_DISK_MAX;i++)
  {
#ifndef DJGPP
	if(param_disk[i]->name)
	  FREE(param_disk[i]->name);
#endif
	FREE(param_disk[i]);    /* forget name */
  }
  return 0;
}

void interface_analyse(t_param_disk param_disk, const int debug)
{
  /* log */
  ecrit_rapport("\nAnalyse ");
  aff_disk_rapport(param_disk);
  /* ncurses interface */
  aff_copy(stdscr);
  mvaddstr(23,0,m_PART_TYPE);
  move(4,0);
  aff_disk(stdscr,param_disk);
  mvaddstr(5,0,"Check current partition structure");
  move(6,0);
  refresh();
  delete_list_part(read_part(param_disk, debug));
  aff_buffer(BUFFER_DISPLAY,"Q",stdscr);
}

int interface_write(t_param_disk param_disk,t_list_part list_part,const int can_search_deeper)
{
  t_list_part parts;
  ecrit_rapport("\ninterface_write()\n");
  aff_copy(stdscr);
  move(4,0);
  aff_disk(stdscr,param_disk);
  move(5,0);
  aff_buffer(BUFFER_RESET,"Q");
  mvaddstr(6,0,m_PART_HEADER_WRITE);
  for(parts=list_part;parts;parts=parts->next)
	if(parts->part->status!=STATUS_LOG)
	  aff_part_buffer(2,param_disk,parts->part);
  for(parts=list_part;parts;parts=parts->next)
	if(parts->part->status==STATUS_LOG)
	  aff_part_buffer(2,param_disk,parts->part);
  return aff_buffer(BUFFER_DISPLAY,can_search_deeper?"SW":"W",stdscr);
}

void interface_options(int *paranoid,int *dump_ind,int *fast_mode,int *align,int *recover_cih, int *allow_partial_last_cylinder,int *ask_part_order,int *halt_on_errors, int *max_ext)
{
/* paranoid: search in partition
 * dump_ind dump sector Y/N
 * fast_mode Partitions are aligned on cylinder boundary Y/N
 * */
  int done = FALSE;
  int menu = 9;
  // ncurses interface
  while (!done) {
    int car;
	struct MenuItem menuOptions[]=
	{
	  { 'P', NULL, "Search all partitions" },
	  { 'D',NULL,"Dump essential sectors" },
	  { 'C',NULL,"Partitions are aligned on cylinder boundary" },
	  { 'F',NULL,"Don't check every hard disk head" },
	  { 'T',NULL,"WARNING: ONLY WORKS WITH  FAT  PARTITION TYPES" },
	  { 'A',NULL,""},
	  { 'O',NULL,""},
	  { 'E',NULL,"Halt on Disk I/O errors" },
	  { 'M',NULL,""},
	  { 'Q',"Quit","Return to main menu"},
	  { 0, NULL, NULL }
	};
	menuOptions[0].name=*paranoid?"Paranoid : Yes":"Paranoid : No";
	menuOptions[1].name=*dump_ind?"Dump : Yes":"Dump : No";
	menuOptions[2].name=*align?"Cylinder boundary : Yes":"Cylinder boundary : No";
	switch(*fast_mode)
	{
	  case 0:
	    menuOptions[3].name="Mode : Fast";
	    break;
	  case 1:
	    menuOptions[3].name="Mode : Medium";
	    break;
	  default:
	    menuOptions[3].name="Mode : Slow";
	    break;
	}
	menuOptions[4].name=*recover_cih?"Recover from CIH virus : Yes":"Recover from CIH virus : No";
	menuOptions[5].name=*allow_partial_last_cylinder?"Allow partial last cylinder : Yes":"Allow partial last cylinder : No";
	menuOptions[6].name=*ask_part_order?"Ask partition order : Yes":"Ask partition order : No";
	menuOptions[7].name=*halt_on_errors?"Halt on errors : Yes":"Halt on errors : No";
	menuOptions[8].name=*max_ext?"Maximize extended part : Yes":"Maximize extended part : No";
	car=toupper( menuSelect(INTER_OPTION_Y, INTER_OPTION_X, menuOptions, 0, "PDCFAOEMQ", MENU_VERT, menu));
	switch(car)
	{
	  case 'P':
		*paranoid=!*paranoid;
		menu=0;
		break;
	  case 'D':
		*dump_ind=!*dump_ind;
		menu=1;
		break;
	  case 'C':
		*align=!*align;
		menu=2;
		break;
	  case 'F':
		if(*fast_mode<2)
		  (*fast_mode)++;
		else
		  *fast_mode=0;
		menu=3;
		break;
	  case 'T':
		*recover_cih=!*recover_cih;
		menu=4;
		break;
	  case 'A':
		*allow_partial_last_cylinder=!*allow_partial_last_cylinder;
		menu=5;
		break;
	  case 'O':
		*ask_part_order=!*ask_part_order;
		menu=6;
		break;
	  case 'E':
		*halt_on_errors=!*halt_on_errors;
		menu=7;
		break;
	  case 'M':
		*max_ext=!*max_ext;
		menu=8;
		break;
	  case ESC:
	  case 'Q':
		done = TRUE;
		break;
	}
  }
  // write new options to log file
  ecrit_rapport("New options :\n Paranoid : %s\n Dump : %s\n Cylinder boudary : %s\n fast_mode=%d(0=fast)\n Allow partial last cylinder : %s\n Ask partition order : %s\n Halt on errors : %s\n Recover from CIH virus : %s\n Maximize extended part : %s\n",
	*paranoid?"Yes":"No",
	*dump_ind?"Yes":"No",
	*align?"Yes":"No",
	*fast_mode,
	*allow_partial_last_cylinder?"Yes":"No",
	*ask_part_order?"Yes":"No",
	*halt_on_errors?"Yes":"No",
	*recover_cih?"Yes":"No",
	*max_ext?"Yes":"No");
}

void dump(WINDOW *window, const void *nom_dump,unsigned int lng)
{
  unsigned int i,j;
  unsigned int pos=0;
  unsigned char car;
  int done=0;
  int menu=2;   /* default : quit */
  struct MenuItem menuDump[]=
  {
    { 'P', "Previous",""},
    { 'N', "Next","" },
    { 'Q',"Quit","Quit dump section"},
    { 0, NULL, NULL }
  };
  /* write dump to log file*/
  for (i=0; (i<lng/0x10); i++)
  {
    ecrit_rapport("%04X ",i*0x10);
    for(j=0; j< 0x10;j++)
    {
      car=*((const unsigned char *)nom_dump+i*0x10+j);
      ecrit_rapport("%02x", car);
      if(j%4==(4-1))
	ecrit_rapport(" ");
    }
    ecrit_rapport("  ");
    for(j=0; j< 0x10;j++)
    {
      car=*((const unsigned char*)nom_dump+i*0x10+j);
      if ((car<32)||(car >= 127))
	ecrit_rapport(".");
      else
	ecrit_rapport("%c",  car);
    }
    ecrit_rapport("\n");
  }
  /* ncurses interface */
  mvwaddstr(window,DUMP_Y,DUMP_X,m_DUMP_HEXA);
  /* On pourrait utiliser wscrl */
  do
  {
    for (i=pos; (i<lng/0x10)&&((i-pos)<DUMP_MAX_LINES); i++)
    {
      wmove(window,DUMP_Y+i-pos,DUMP_X);
      wclrtoeol(window);
      wdoprintf(window,"%04X ",i*0x10);
      for(j=0; j< 0x10;j++)
      {
	car=*((const unsigned char*)nom_dump+i*0x10+j);
	wdoprintf(window,"%02x", car);
	if(j%4==(4-1))
	  wdoprintf(window," ");
      }
      wdoprintf(window,"  ");
      for(j=0; j< 0x10;j++)
      {
	car=*((const unsigned char*)nom_dump+i*0x10+j);
	if ((car<32)||(car >= 127))
	  wdoprintf(window,".");
	else
	  wdoprintf(window,"%c",  car);
      }
    }
    switch (wmenuSelect(window,INTER_DUMP_Y, INTER_DUMP_X, menuDump, 8, "PNQ", MENU_HORIZ | MENU_ACCEPT_OTHERS, menu))
    {
      case 'p':
      case 'P':
      case MENU_UP:
	menu=0;
	if(pos>0)
	  pos--;
	break;
      case 'n':
      case 'N':
      case MENU_DOWN:
	menu=1;
	if(pos<lng/0x10-DUMP_MAX_LINES)
	  pos++;
	break;
      case KEY_PPAGE:
	menu=0;
	if(pos>DUMP_MAX_LINES-1)
	  pos-=DUMP_MAX_LINES-1;
	else
	  pos=0;
	break;
      case KEY_NPAGE:
	menu=1;
	if(pos<lng/0x10-DUMP_MAX_LINES-(DUMP_MAX_LINES-1))
	  pos+=DUMP_MAX_LINES-1;
	else
	  pos=lng/0x10-DUMP_MAX_LINES;
	break;
      case ESC:
      case 'q':
      case 'Q':
	done = TRUE;
	break;
    }
  } while(!done);
}

void dump2(WINDOW *window, const void *dump_1, const void *dump_2, const unsigned int lng)
{
  unsigned int i,j;
  unsigned int pos=0;
  int done=0;
  int menu=2;   /* default : quit */
  struct MenuItem menuDump[]=
  {
    { 'P', "Previous",""},
    { 'N', "Next","" },
    { 'Q',"Quit","Quit dump section"},
    { 0, NULL, NULL }
  };
  /* write dump to log file*/
  for (i=0; (i<lng/0x8); i++)
  {
    ecrit_rapport("%04X ",i*0x08);
    for(j=0; j< 0x08;j++)
    {
      unsigned char car=*((const unsigned char *)dump_1+i*0x08+j);
      ecrit_rapport("%02x", car);
      if(j%4==(4-1))
	ecrit_rapport(" ");
    }
    ecrit_rapport("  ");
    for(j=0; j< 0x08;j++)
    {
      unsigned char car=*((const unsigned char*)dump_1+i*0x08+j);
      if ((car<32)||(car >= 127))
	ecrit_rapport(".");
      else
	ecrit_rapport("%c",  car);
    }
    ecrit_rapport("  ");
    for(j=0; j< 0x08;j++)
    {
      unsigned char car=*((const unsigned char *)dump_2+i*0x08+j);
      ecrit_rapport("%02x", car);
      if(j%4==(4-1))
	ecrit_rapport(" ");
    }
    ecrit_rapport("  ");
    for(j=0; j< 0x08;j++)
    {
      unsigned char car=*((const unsigned char*)dump_2+i*0x08+j);
      if ((car<32)||(car >= 127))
	ecrit_rapport(".");
      else
	ecrit_rapport("%c",  car);
    }
    ecrit_rapport("\n");
  }
  /* ncurses interface */
  do
  {
    for (i=pos; (i<lng/0x08)&&((i-pos)<DUMP_MAX_LINES); i++)
    {
      wmove(window,DUMP_Y+i-pos,DUMP_X);
      wclrtoeol(window);
      wdoprintf(window,"%04X ",i*0x08);
      for(j=0; j< 0x08;j++)
      {
	unsigned char car=*((const unsigned char*)dump_1+i*0x08+j);
	wdoprintf(window,"%02x", car);
	if(j%4==(4-1))
	  wdoprintf(window," ");
      }
      wdoprintf(window,"  ");
      for(j=0; j< 0x08;j++)
      {
	unsigned char car=*((const unsigned char*)dump_1+i*0x08+j);
	if ((car<32)||(car >= 127))
	  wdoprintf(window,".");
	else
	  wdoprintf(window,"%c",  car);
      }
      wdoprintf(window,"  ");
      for(j=0; j< 0x08;j++)
      {
	unsigned char car=*((const unsigned char*)dump_2+i*0x08+j);
	wdoprintf(window,"%02x", car);
	if(j%4==(4-1))
	  wdoprintf(window," ");
      }
      wdoprintf(window,"  ");
      for(j=0; j< 0x08;j++)
      {
	unsigned char car=*((const unsigned char*)dump_2+i*0x08+j);
	if ((car<32)||(car >= 127))
	  wdoprintf(window,".");
	else
	  wdoprintf(window,"%c",  car);
      }
    }
    switch (wmenuSelect(window,INTER_DUMP_Y, INTER_DUMP_X, menuDump, 8, "PNQ", MENU_HORIZ | MENU_ACCEPT_OTHERS, menu))
    {
      case 'p':
      case 'P':
      case MENU_UP:
	menu=0;
	if(pos>0)
	  pos--;
	break;
      case 'n':
      case 'N':
      case MENU_DOWN:
	menu=1;
	if(pos<lng/0x08-DUMP_MAX_LINES)
	  pos++;
	break;
      case KEY_PPAGE:
	menu=0;
	if(pos>DUMP_MAX_LINES-1)
	  pos-=DUMP_MAX_LINES-1;
	else
	  pos=0;
	break;
      case KEY_NPAGE:
	menu=1;
	if(pos<lng/0x08-DUMP_MAX_LINES-(DUMP_MAX_LINES-1))
	  pos+=DUMP_MAX_LINES-1;
	else
	  pos=lng/0x08-DUMP_MAX_LINES;
	break;
      case ESC:
      case 'q':
      case 'Q':
	done = TRUE;
	break;
    }
  } while(!done);
}

int dump_editor(const unsigned char *nom_dump,const unsigned int lng, const int menu_pos)
{
  unsigned int i,j;
  unsigned int pos;
  unsigned char car;
  int done=0;
  int menu;
  struct MenuItem menuDump[]=
  {
	{ 'P', "Previous",""},
	{ 'N', "Next","" },
	{ 'Q',"Quit","Quit dump section"},
	{ 0, NULL, NULL }
  };
  /* write dump to log file*/
  for (i=0; (i<lng/0x10); i++)
  {
	ecrit_rapport("%04X ",i*0x10);
	for(j=0; j< 0x10;j++)
	{
	  car=*(nom_dump+i*0x10+j);
	  ecrit_rapport("%02x", car);
	  if(j%4==(4-1))
		ecrit_rapport(" ");
	}
	ecrit_rapport("  ");
	for(j=0; j< 0x10;j++)
	{
	  car=*(nom_dump+i*0x10+j);
	  if ((car<32)||(car >= 127))
		ecrit_rapport(".");
	  else
		ecrit_rapport("%c",  car);
	}
	ecrit_rapport("\n");
  }
  /* ncurses interface */
  pos=(menu_pos==MENU_DOWN?0:lng/0x10-DUMP_MAX_LINES);
  menu=(menu_pos==MENU_DOWN?1:0);
  mvaddstr(DUMP_Y,DUMP_X,m_DUMP_HEXA);
  do
  {
	for (i=pos; (i<lng/0x10)&&((i-pos)<DUMP_MAX_LINES); i++)
	{
	  move(DUMP_Y+i-pos,DUMP_X);
	  clrtoeol();
	  doprintf("%04X ",i*0x10);
	  for(j=0; j< 0x10;j++)
	  {
		car=*(nom_dump+i*0x10+j);
		doprintf("%02x", car);
		if(j%4==(4-1))
		  doprintf(" ");
	  }
	  doprintf("  ");
	  for(j=0; j< 0x10;j++)
	  {
		car=*(nom_dump+i*0x10+j);
		if ((car<32)||(car >= 127))
		  doprintf(".");
		else
		  doprintf("%c",  car);
	  }
	}
	switch (menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuDump, 8, "PNQ", MENU_HORIZ | MENU_ACCEPT_OTHERS, menu))
	{
	  case 'p':
	  case 'P':
	  case MENU_UP:
		menu=0;
		if(pos>0)
		  pos--;
		else
		  done=MENU_UP;
		break;
	  case 'n':
	  case 'N':
	  case MENU_DOWN:
		menu=1;
		if(pos<lng/0x10-DUMP_MAX_LINES)
		  pos++;
		else
		  done = MENU_DOWN;
		break;
	  case KEY_PPAGE:
		menu=0;
		if(pos==0)
		  done=MENU_UP;
		if(pos>DUMP_MAX_LINES-1)
		  pos-=DUMP_MAX_LINES-1;
		else
		  pos=0;
		break;
	  case KEY_NPAGE:
		menu=1;
		if(pos==lng/0x10-DUMP_MAX_LINES)
		  done=MENU_DOWN;
		if(pos<lng/0x10-DUMP_MAX_LINES-(DUMP_MAX_LINES-1))
		  pos+=DUMP_MAX_LINES-1;
		else
		  pos=lng/0x10-DUMP_MAX_LINES;
		break;
	  case ESC:
	  case 'q':
	  case 'Q':
		done = 'Q';
		break;
	}
  } while(!done);
  return done;
}

int aff_buffer(const int cmd, const char *_format, ...)
{
  static char buffer[MAX_LINES][LINE_LENGTH+1];
  static int nbr_lines=0;
  va_list ap;
  va_start(ap,_format);
  switch(cmd)
  {
	case BUFFER_RESET:
	  {
		int i;
		nbr_lines=0;
		for(i=0;i<MAX_LINES;i++)
		  memset(buffer[i],0,LINE_LENGTH+1);
	  }
	  break;
	case BUFFER_ADD:
	  {
		char tmp_line[BUFFER_LINE_LENGTH+1];
		char *pos_in_tmp_line=tmp_line-1;
		memset(tmp_line,'\0',sizeof(tmp_line));
#ifdef DJGPP
		if(strlen(_format)>sizeof(tmp_line))
		  ecrit_rapport("\nBUG: %s\n",_format);
		vsprintf(tmp_line,_format,ap);
#else
		vsnprintf(tmp_line,BUFFER_LINE_LENGTH,_format,ap);
#endif
		while(pos_in_tmp_line && (nbr_lines<MAX_LINES))
		{
		  char *ret_ligne= strchr(++pos_in_tmp_line,'\n');
		  if(ret_ligne)
		  {
			strncat(buffer[nbr_lines],pos_in_tmp_line,
				ret_ligne-pos_in_tmp_line<LINE_LENGTH-strlen(buffer[nbr_lines])?ret_ligne-pos_in_tmp_line:LINE_LENGTH-strlen(buffer[nbr_lines]));
			if(strlen(buffer[nbr_lines])>0)
			  nbr_lines++;
		  }
		  else
		  {
			strncat(buffer[nbr_lines],pos_in_tmp_line,LINE_LENGTH-strlen(buffer[nbr_lines]));
		  }
		  pos_in_tmp_line=ret_ligne;
		}
//		ecrit_rapport("aff_buffer %d =>%s<=\n",nbr_lines,tmp_line);
		if(nbr_lines>=MAX_LINES)
		{
		  ecrit_rapport("aff_buffer too much lines =>%s<=\n",tmp_line);
		}
	  }
	  break;
	case BUFFER_DISPLAY:
	  {
		int i;
		int menu;
		int pos=0;
		int done=0;
		WINDOW *window=*(WINDOW**)ap;
		struct MenuItem menuBuffer[]=
		{
		  { 'P', "Previous",""},
		  { 'N', "Next","" },
		  { 'Q',"Quit","Quit this section"},
		  { 'S',"Search!","Search deeper, try to find more partitions"},
		  { 'W', "Write","Write partition structure to disk"},
		  { 'O', "Org. BS","Copy boot sector over backup sector"},
		  { 'B', "Backup BS","Copy backup boot sector over boot sector"},
		  { 'R', "Rebuild BS","Rebuild boot sector"},
		  { 'D', "Dump","Dump boot sector and backup boot sector"},
		  { 0, NULL, NULL }
		};
		char options[20];
		if(nbr_lines<=DUMP_MAX_LINES)
		{
		  strcpy(options,"Q");
		  menu=0;       /* default = quit */
		}
		else
		{
		  strcpy(options,"PNQ");
		  menu=1;       /* default = next */
		}
		strncat(options,_format,sizeof(options)-strlen(options));
		if(buffer[nbr_lines][0])
		  nbr_lines++;
		/* to log file */
		for(i=0;i<nbr_lines;i++)
		  ecrit_rapport("%s\n",buffer[i]);
		/* curses interface */
		do
		{
		  for (i=pos; (i<nbr_lines)&&((i-pos)<DUMP_MAX_LINES); i++)
		  {
			wmove(window,DUMP_Y+i-pos,DUMP_X);
			wclrtoeol(window);
			wdoprintf(window,"%s",buffer[i]);
		  }
		  
		  switch (wmenuSelect(window,INTER_DUMP_Y, INTER_DUMP_X, menuBuffer, 8, options, MENU_HORIZ | MENU_ACCEPT_OTHERS, menu))
		  {
			case 'p':
			case 'P':
			case MENU_UP:
			  menu=0;
			  if(pos>0)
				pos--;
			  break;
			case 'n':
			case 'N':
			case MENU_DOWN:
			  menu=1;
			  if(pos<nbr_lines-DUMP_MAX_LINES)
				pos++;
			  break;
			case KEY_PPAGE:
			  menu=0;
			  if(pos>DUMP_MAX_LINES-1)
			    pos-=DUMP_MAX_LINES-1;
			  else
			    pos=0;
			  break;
			case KEY_NPAGE:
			  menu=1;
			  if(pos<nbr_lines-DUMP_MAX_LINES-(DUMP_MAX_LINES)-1)
			    pos+=DUMP_MAX_LINES-1;
			  else
			    pos=nbr_lines-DUMP_MAX_LINES>0?nbr_lines-DUMP_MAX_LINES:0;
			  break;
			case 's':
			case 'S':
			  if(strchr(options,'S'))
				return 2;
			  break;
			case 'o':
			case 'O':
			  if(strchr(options,'O'))
				return 1;
			  break;
			case 'b':
			case 'B':
			  if(strchr(options,'B'))
				return 2;
			  break;
			case 'r':
			case 'R':
			  if(strchr(options,'R'))
				return 3;
			  break;
			case 'd':
			case 'D':
			  if(strchr(options,'D'))
				return 4;
			  break;
			case 'w':
			case 'W':
			  if(strchr(options,'W'))
				return 1;
			  break;
			case ESC:
			case 'q':
			case 'Q':
			  done = TRUE;
			  break;
		  }
		} while(!done);
	  }
	  break;
	case BUFFER_SHOW:
	  {
	    int i;
	    int pos=nbr_lines-DUMP_MAX_LINES<0?0:nbr_lines-DUMP_MAX_LINES;
	    if(buffer[nbr_lines][0])
	      nbr_lines++;
	    /* curses interface */
	    for (i=pos; (i<nbr_lines)&&((i-pos)<DUMP_MAX_LINES); i++)
	    {
	      move(DUMP_Y+1+i-pos,DUMP_X);
	      clrtoeol();
	      doprintf("%s",buffer[i]);
	    }
	    refresh();
	  }
	  break;
	case BUFFER_WRITE:
	  {
	    int i;
	    if(buffer[nbr_lines][0])
	      nbr_lines++;
	    /* to log file */
	    for(i=0;i<nbr_lines;i++)
	    {
	      printf("%s\n",buffer[i]);
	      ecrit_rapport("%s\n",buffer[i]);
	    }
	  }
	  break;
  }
  va_end(ap);
  return 0;
}

/*
part_num  2
PLEX      1
part_type ...
CHS 5+1+3+1+2=12
size	  9
part_name EXT2 16, FAT 11

part_num PLEX part_type CHS CHS size [part_name]
2       1  1 1       x 1 121 121  9 11      16 1
2+1+1+1+x+1+12+1+12+1+9+1+1+16+1=60+x
 * */

void aff_part_buffer(const int newline,const t_param_disk disk_car,const t_diskext partition)
{
  switch(newline)
  {
    case 1:
      aff_buffer(BUFFER_ADD,"\n   ");
      break;
    case 2:
      if((partition->status!=STATUS_EXT_IN_EXT) && partition->order)
	aff_buffer(BUFFER_ADD,"%2d ", partition->order);
      else
	aff_buffer(BUFFER_ADD,"   ");
      break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           aff_buffer(BUFFER_ADD,"P"); break;
    case STATUS_PRIM_BOOT:      aff_buffer(BUFFER_ADD,"*"); break;
    case STATUS_EXT:            aff_buffer(BUFFER_ADD,"E"); break;
    case STATUS_EXT_IN_EXT:     aff_buffer(BUFFER_ADD,"X"); break;
    case STATUS_LOG:            aff_buffer(BUFFER_ADD,"L"); break;
    case STATUS_DELETED:        aff_buffer(BUFFER_ADD,"D"); break;
  }
  if(nom_os[partition->part_type])
    aff_buffer(BUFFER_ADD," %-20s ",nom_os[partition->part_type]);
  else
    aff_buffer(BUFFER_ADD," Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_from_LBA_buffer(disk_car,partition->lba);
  aff_CHS_from_LBA_buffer(disk_car,partition->lba+partition->part_size-1);
  aff_buffer(BUFFER_ADD,"%10lu", partition->part_size);
  if(partition->name[0])
    aff_buffer(BUFFER_ADD," [%s]",partition->name);
  if(newline==2)
    aff_buffer(BUFFER_ADD,"\n");
}

void aff_entree(const t_entree entree)
{
  if(nom_os[os(entree)])
    doprintf("\n %-20s ",nom_os[os(entree)]);
  else
    doprintf("\n Sys=%02X\t\t\t  ", os(entree));
  doprintf("%4u %3u %2u"
	 " %4u %3u %2u"
	 " %10lu"
	 " %10lu",
	  s_cyl(entree),s_head(entree),s_sect(entree),
	  e_cyl(entree),e_head(entree),e_sect(entree),
	  sect_rel(entree),nbr_sect(entree));
}

void aff_entree_rapport(const t_entree entree)
{
  if(nom_os[os(entree)])
    ecrit_rapport(" %-20s ",nom_os[os(entree)]);
  else
    ecrit_rapport(" Sys=%02X\t\t\t  ", os(entree));
  ecrit_rapport("%4u %3u %2u"
	 " %4u %3u %2u"
	 " %10lu"
	 " %10lu\n",
	  s_cyl(entree),s_head(entree),s_sect(entree),
	  e_cyl(entree),e_head(entree),e_sect(entree),
	  sect_rel(entree),nbr_sect(entree));
}

void aff_CHS_rapport(const t_CHS * CHS)
{
  ecrit_rapport("%5u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS(const t_CHS * CHS)
{
  doprintf("%5u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS_buffer(const t_CHS * CHS)
{
  aff_buffer(BUFFER_ADD,"%5u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS_from_LBA(WINDOW *window, const t_param_disk disk_car, dword hd_offset)
{
  wdoprintf(window,"%5u %3u %2u ", LBA2cylinder(disk_car,hd_offset), LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
}

void aff_CHS_from_LBA_buffer(const t_param_disk disk_car, dword hd_offset)
{
  aff_buffer(BUFFER_ADD,"%5u %3u %2u ", LBA2cylinder(disk_car,hd_offset), LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
}

void aff_CHS_from_LBA_rapport(const t_param_disk disk_car, dword hd_offset)
{
  ecrit_rapport("%5u %3u %2u ", LBA2cylinder(disk_car,hd_offset), LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
}

void aff_part_rapport(const int newline,const t_param_disk disk_car,const t_diskext partition)
{
  switch(newline)
  {
    case 1:
      ecrit_rapport("\n   ");
      break;
    case 2: 
      if((partition->status!=STATUS_EXT_IN_EXT) && partition->order)
	ecrit_rapport("%2d ", partition->order);
      else
	ecrit_rapport("   ");
      break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           ecrit_rapport("P"); break;
    case STATUS_PRIM_BOOT:      ecrit_rapport("*"); break;
    case STATUS_EXT:            ecrit_rapport("E"); break;
    case STATUS_EXT_IN_EXT:     ecrit_rapport("X"); break;
    case STATUS_LOG:            ecrit_rapport("L"); break;
    case STATUS_DELETED:        ecrit_rapport("D"); break;
  }
  if(nom_os[partition->part_type])
    ecrit_rapport(" %-20s ",nom_os[partition->part_type]);
  else
    ecrit_rapport(" Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_from_LBA_rapport(disk_car,partition->lba);
  aff_CHS_from_LBA_rapport(disk_car,partition->lba+partition->part_size-1);
  ecrit_rapport("%10lu", partition->part_size);
  if(partition->name[0])
    ecrit_rapport(" [%s]",partition->name);
  if(newline==2)
    ecrit_rapport("\n");
}

void aff_part(WINDOW *window,const int newline,const t_param_disk disk_car,const t_diskext partition)
{
  switch(newline)
  {
    case 1:
      wdoprintf(window,"\n   ");
      break;
    case 2:
      if(partition->status!=STATUS_EXT_IN_EXT)
	wdoprintf(window,"%2d ", partition->order);
      else
	wdoprintf(window,"   ");
      break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           wdoprintf(window,"P"); break;
    case STATUS_PRIM_BOOT:      wdoprintf(window,"*"); break;
    case STATUS_EXT:            wdoprintf(window,"E"); break;
    case STATUS_EXT_IN_EXT:     wdoprintf(window,"X"); break;
    case STATUS_LOG:            wdoprintf(window,"L"); break;
    case STATUS_DELETED:        wdoprintf(window,"D"); break;
  }
  if(nom_os[partition->part_type])
    wdoprintf(window," %-20s ",nom_os[partition->part_type]);
  else
    wdoprintf(window," Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_from_LBA(window,disk_car,partition->lba);
  aff_CHS_from_LBA(window,disk_car,partition->lba+partition->part_size-1);
  wdoprintf(window,"%10lu", partition->part_size);
  if(partition->name[0])
    wdoprintf(window," [%s]",partition->name);
}

void aff_LBA2CHS_rapport(const t_param_disk disk_car, const dword pos_LBA)
{
  dword tmp;
  dword cylinder, head, sector;
  tmp=disk_car->CHS.sector;
  sector=(pos_LBA%tmp)+1;
  tmp=pos_LBA/tmp;
  cylinder=tmp/(disk_car->CHS.head+1);
  head=tmp%(disk_car->CHS.head+1);
  ecrit_rapport("%ld/%ld/%ld", cylinder,head,sector);
}

void aff_disk(WINDOW *window,const t_param_disk param_disk)
{
#ifdef DJGPP
  wdoprintf(window, m_DISK " %02x - CHS %d %d %d - %d MB (%s BIOS mode)\n",
      param_disk->disk, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024,param_disk->mode==0?"Std":param_disk->mode==1?"Enh":"Mxd");
#else
  wdoprintf(window, m_DISK " %s - CHS %d %d %d - %d MB\n",
      param_disk->name, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#endif
}

void aff_disk_classic(const t_param_disk param_disk)
{
#ifdef DJGPP
  printf(m_DISK " %02x - CHS %d %d %d - %d MB (%s BIOS mode)\n",
      param_disk->disk, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024,param_disk->mode==0?"Std":param_disk->mode==1?"Enh":"Mxd");
#else
  printf(m_DISK " %s - CHS %d %d %d - %d MB\n",
      param_disk->name, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#endif
}

void aff_disk_rapport(const t_param_disk param_disk)
{
#ifdef DJGPP
  ecrit_rapport(m_DISK " %02x - CHS %d %d %d - %d MB (%s BIOS mode)\n",
      param_disk->disk, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024,param_disk->mode?"Enh":"Std");
#else
  ecrit_rapport(m_DISK " %s - CHS %d %d %d - %d MB\n",
      param_disk->name, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#endif
}

int change_part_type(const t_param_disk disk_car,t_diskext partition)
{
  char response[100];
  int nbr=0;
  int i;
  aff_buffer(BUFFER_RESET,"Q");
  for(i=0;i<=255;i++)
  {
    if(nom_os[i] && strcmp(nom_os[i],"reserved"))
    {
      aff_buffer(BUFFER_ADD,"%02x %-20s ",i,nom_os[i]);
      nbr++;
      if(nbr==3)
      {
	nbr=0;
	aff_buffer(BUFFER_ADD,"\n");
      }
    }
  }
  aff_copy(stdscr);
  move(4,0);
  aff_part(stdscr,0,disk_car,partition);
  aff_buffer(BUFFER_DISPLAY,"Q",stdscr);
  move(23,0);
  doprintf("New partition type [current %02X] ?",partition->part_type);
  if (get_string(response, sizeof(response), NULL) > 0) {
    int tmp_val = strtol(response, NULL, 16);
    if (tmp_val > 0 && tmp_val <= 255)
      partition->part_type = tmp_val;
  }
  return 0;
}

t_list_part ask_structure(const t_param_disk disk_car,t_list_part list_part, const int debug)
{
  int quit=0;
  int offset=0;
  int pos_num=0;
  t_list_part pos=list_part;
  int rewrite=1;
  do
  {
    int i;
    t_list_part parts;
    int car;
    if(rewrite)
    {
      aff_copy(stdscr);
      move(4,0);
      aff_disk(stdscr,disk_car);
      mvaddstr(23,0,m_PART_TYPE);
      mvaddstr(5,0,m_CHOOSE_PART_TYPE);
      mvaddstr(6,0,m_CHOOSE_PART_TYPE_END);
      mvaddstr(7,0,m_PART_HEADER2);
      rewrite=0;
    }
    for(i=0,parts=list_part;parts && (i<offset);parts=parts->next,i++);
    for(i=offset;parts &&((i-offset)<INTER_STRUCTURE);i++,parts=parts->next)
    {
      move(8+i-offset,0);
      clrtoeol();	/* before addstr for BSD compatibility */
      if(parts==pos)
      {
	standout();
	aff_part(stdscr,0,disk_car,parts->part);
	standend();
	move(24,0);
	clrtoeol();	/* before addstr for BSD compatibility */
	addstr(parts->part->info);
      } else
      {
	aff_part(stdscr,0,disk_car,parts->part);
      }
    }
    if(test_structure(list_part))
      mvaddstr(22,0,m_STRUCT_BAD);
    else
      mvaddstr(22,0,m_STRUCT_OK);
    move(22,15);
    clrtoeol();	/* before addstr for BSD compatibility */
    addstr("Keys a: add partition");
    if(list_part)
    {
      addstr(", T: change type");
      switch(pos->part->upart_type)
      {
	case UP_EXT2:
	case UP_EXT3:
	case UP_RFS:
	case UP_RFS2:
	case UP_FAT12:
	case UP_FAT16:
	case UP_FAT32:
//	case UP_NTFS:
	  addstr(", P: list files");
	  break;
      }
    }
    refresh();
    car=wgetch(stdscr);
    switch(car)
    {
      case 'q':
      case '\r':
      case '\n':
      case KEY_ENTER:
#ifdef DJGPP
      case PADENTER:
#endif
      case 'M':
	quit=1;
	break;
    }
    switch(car)
    {
      case KEY_UP:
	if(list_part)
	{
	  only_one_bootable(list_part,pos);
	  if(pos->prev)
	  {
	    pos=pos->prev;
	    pos_num--;
	  }
	  if(pos_num<offset)
	    offset--;
	}
	break;
      case KEY_DOWN:
	if(list_part)
	{
	  only_one_bootable(list_part,pos);
	  if(pos->next)
	  {
	    pos=pos->next;
	    pos_num++;
	  }
	  if(pos_num>=offset+INTER_STRUCTURE)
	    offset++;
	}
	break;
      case KEY_PPAGE:
	if(list_part)
	{
	  only_one_bootable(list_part,pos);
	  for(i=0;(i<INTER_STRUCTURE) && (pos->prev!=NULL);i++)
	  {
	    pos=pos->prev;
	    pos_num--;
	    if(pos_num<offset)
	      offset--;
	  }
	}
	break;
      case KEY_NPAGE:
	if(list_part)
	{
	  only_one_bootable(list_part,pos);
	  for(i=0;(i<INTER_STRUCTURE) && (pos->next!=NULL);i++)
	  {
	    pos=pos->next;
	    pos_num++;
	    if(pos_num>=offset+INTER_STRUCTURE)
	      offset++;
	  }
	}
	break;
      case KEY_RIGHT:
      case '+':
      case ' ':
	if(list_part)
	{
	  if(pos->part->status==STATUS_LOG)
	    pos->part->status=STATUS_DELETED;
	  else
	    pos->part->status++;
	  if(pos->part->status==STATUS_LOG && !can_be_ext(disk_car,pos->part))
	    pos->part->status=STATUS_DELETED;
	}
	break;
      case KEY_LEFT:
      case '-':
	if(list_part)
	{
	  if(pos->part->status==STATUS_DELETED)
	    pos->part->status=STATUS_LOG;
	  else
	    pos->part->status--;
	  if(pos->part->status==STATUS_LOG && !can_be_ext(disk_car,pos->part))
	    pos->part->status--;
	}
	break;
      case 'a':
      case 'A':
	{
	  list_part=add_partition(disk_car,list_part, debug);
	  rewrite=1;
	  offset=0;
	  pos_num=0;
	  pos=list_part;
	}
	break;
      case 't':
      case 'T':
	if(list_part)
	{
	  rewrite=1;
	  change_part_type(disk_car,pos->part);
	}
	break;
      case 'p':
      case 'P':
	if(list_part)
	{
	  WINDOW *window=newwin(0,0,0,0);	/* full screen */
	  aff_copy(window);
	  switch(pos->part->upart_type)
	  {
	    case UP_FAT12:
	    case UP_FAT16:
	    case UP_FAT32:
	      dump_fat_root(window,disk_car,pos->part,debug);
	      break;
	    case UP_EXT2:
	    case UP_EXT3:
	      dir_partition_ext2(window,disk_car,pos->part,debug);
	      break;
	    case UP_RFS:
	    case UP_RFS2:
	      dir_partition_reiser(window,disk_car,pos->part,debug);
	      break;
	    case UP_NTFS:
	      dir_ntfs(window,disk_car,pos->part);
	      break;
	  }
	  delwin(window);
#ifdef DJGPP
	  wredrawln(stdscr,0,stdscr->_maxy);	/* redrawwin def is boggus in pdcur24 */
#else
	  redrawwin(stdscr);	/* stdscr has been corrupted by window */
#endif
	}
	break;
      default:
//	ecrit_rapport("ask_structure car=%x\n",car);
	break;
    }
  } while(!quit);
  return list_part;
}

void interface_editor(t_param_disk param_disk)
{
  int done = FALSE;
  dword hd_offset=0;
  aff_disk_rapport(param_disk);
  while (!done)
  {
    static struct MenuItem menuEditor[]=
    {
      { 'c', "Change position", "" },
      { 'd', "Dump", "Dump sector" },
      { 'q', "Quit",""},
      { 0, NULL, NULL }
    };
    switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuEditor, 8, "cdq", MENU_HORIZ | MENU_ACCEPT_OTHERS, 0)))
    {
      case 'C':
	interface_editor_position(param_disk,&hd_offset);
	break;
      case 'D':
	{
	  unsigned char buffer[0x200];
	  int menu_pos=MENU_DOWN;
	  while(!done)
	  {
	    move(4,0);
	    clrtoeol();
	    doprintf("%ld", hd_offset);
	    if(hd_read_LBA(param_disk,1, buffer, hd_offset))
	      doprintf(m_PART_RD_ERR);
	    else
	    {
	      menu_pos=dump_editor(buffer,0x200,menu_pos);
	      switch(menu_pos)
	      {
		case MENU_UP:
		  if(hd_offset>0)
		    hd_offset--;
		  else
		    menu_pos=MENU_DOWN;
		  break;
		case MENU_DOWN:
		  if(hd_offset<param_disk->lba-1)
		    hd_offset++;
		  else
		    menu_pos=MENU_UP;
		  break;
		default:
		  done = TRUE;
		  break;
	      }
	    }
	  }
	}
	break;
      case ESC:
      case 'Q':
	done = TRUE;
	break;
    }
  }
}

void interface_editor_position(const t_param_disk disk_car,unsigned long int *lba)
{
  t_CHS position;
  int done = FALSE;
  char def[LINE_LENGTH];
  char response[LINE_LENGTH];
  unsigned int tmp_val;
  int command;
  position.cylinder=LBA2cylinder(disk_car,*lba);
  position.head=LBA2head(disk_car,*lba);
  position.sector=LBA2sector(disk_car,*lba);
  while (!done) {
	static struct MenuItem menuGeometry[]=
	{
	  { 'c', "Cylinders", "Change cylinder" },
	  { 'h', "Heads", "Change head" },
	  { 's', "Sectors", "Change sector" },
	  { 'd', "Done", "Done with changing" },
	  { 0, NULL, NULL }
	};
	move(COMMAND_LINE_Y, COMMAND_LINE_X);
	clrtoeol();
	refresh();
	command=menuSimple(menuGeometry, 3);
	command=toupper(command);
	switch (command) {
	  case 'C':
		sprintf(def, "%d", position.cylinder);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of cylinders: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val <= disk_car->CHS.cylinder) {
			position.cylinder = tmp_val;
		  } else
			doprintf("Illegal cylinders value");
		}
		break;
	  case 'H':
		sprintf(def, "%d", position.head);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of heads: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val <= disk_car->CHS.head) {
			position.head = tmp_val;
		  } else
			doprintf("Illegal heads value");
		}
		break;
	  case 'S':
		sprintf(def, "%d", position.sector);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of sectors per track: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= disk_car->CHS.sector ) {
			position.sector = tmp_val;
		  } else
			doprintf("Illegal sectors value");
		}
		break;
	  case ESC:
	  case 'D':
		done = TRUE;
		break;
	}
  }
  *lba=CHS2LBA(disk_car,&position);
}

void interface_list(t_param_disk disk_car, const int debug,const int test_recovery)
{ /* log */
  t_list_part list_part;
  ecrit_rapport("\nAnalyse ");
  aff_disk_rapport(disk_car);
  aff_disk_classic(disk_car);
  list_part=read_part(disk_car, debug);
  aff_buffer(BUFFER_WRITE,"Q");
  if(test_recovery)
  {
    t_list_part element;
    ecrit_rapport("rebuild_FAT_BS()\n");
    for(element=list_part;element;element=element->next)
    {
      if(is_fat(element->part->part_type))		/* do not use upart_type */
      {
	aff_part_rapport(2,disk_car,element->part);
	rebuild_FAT_BS(disk_car,element->part,debug,0,0);	/* dump_ind */
      }
    }
  }
  delete_list_part(list_part);
}

int ask_confirmation(const char*msg)
{
  int res;
  WINDOW *window=newwin(0,0,0,0);	/* full screen */
  aff_copy(window);
  res=ask_YN(window,msg);
  delwin(window);
#ifdef DJGPP
  wredrawln(stdscr,0,stdscr->_maxy);	/* redrawwin def is boggus in pdcur24 */
#else
  redrawwin(stdscr);	/* stdscr has been corrupted by window */
#endif
  return res;
}
