/*

    File: ntfs.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
//#define NTFS_DEBUG 1

#define NTFS_GETU8(p)      (*(const __u8*)(p))
#define NTFS_GETU16(p)     (*(const __u16*)(p))
#define NTFS_GETU32(p)     (*(const __u32*)(p))
#define NTFS_GETU64(p)     (*(const __u64*)(p))
 
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "types.h"
#include "common.h"
#include "intrface.h"
#include "ntfs.h"
#include "fnctdsk.h"
#include "lang.h"
#include "hdaccess.h"
int create_ntfs_boot_sector(t_param_disk disk_car,const t_diskext partition, const int part_type, const int interface, const __u8 cluster_size, const __u64 mft_cluster, const __u64 mftbak_cluster, const __s32 clusters_frs, const __s32 clusters_record);
int dump_ntfs_info(const struct ntfs_boot_sector *ntfs_header);
int dump_2ntfs_info(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2);
int dump_2ntfs_rapport(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2);

int ntfs_read_MFT(t_param_disk disk_car, t_diskext partition, const struct ntfs_boot_sector*ntfs_header,const int my_type);

int ntfs_get_attr(const char *attr, const int my_type, t_diskext partition, const char *end);

int check_NTFS(t_param_disk disk_car,t_diskext partition,const int debug,const int dump_ind)
{
  unsigned char buffer[0x200];
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba)!=0)
  { return 1; }
  if(test_NTFS(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind)!=0)
    return 1;
  set_NTFS_info(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind);
  return 0;
}

int recover_NTFS(t_param_disk disk_car, const struct ntfs_boot_sector*ntfs_header,t_diskext partition,const int debug, const int dump_ind, const int backup)
{
  if(test_NTFS(disk_car,ntfs_header,partition,debug,dump_ind)!=0)
    return 1;
  partition->part_size=ntfs_header->sectors_nbr+1;
  partition->part_type=(unsigned char)P_NTFS;
  if(backup)
  {
    partition->boot_sector=partition->part_size-1;
    partition->lba-=partition->part_size-1;
  }
  set_NTFS_info(disk_car,ntfs_header,partition,debug,dump_ind);
  return 0;
}

int set_NTFS_info(t_param_disk disk_car, const struct ntfs_boot_sector*ntfs_header,t_diskext partition,const int debug, const int dump_ind)
{
  partition->name[0]='\0';
  return ntfs_read_MFT(disk_car, partition, ntfs_header,0x60);
}

int test_NTFS(const t_param_disk disk_car,const struct ntfs_boot_sector*ntfs_header, t_diskext partition,const int debug, const int dump_ind)
{
  const char*buffer=(const char*)ntfs_header;
  if(ntfs_header->marker==0xAA55)
  {
    if(memcmp(ntfs_header->system_id,"NTFS",4)==0)
    {
      if((ntfs_header->reserved>0) || (ntfs_header->fats>0) ||
	  (ntfs_header->dir_entries[0]!=0) || (ntfs_header->dir_entries[1]!=0) ||
	  (ntfs_header->sectors[0]!=0) || (ntfs_header->sectors[1]!=0) ||
	  (ntfs_header->fat_length!=0) || (ntfs_header->total_sect!=0) || (ntfs_header->cluster_size==0))
	return 1;
      if((debug!=0) || (dump_ind!=0))
      {
	ecrit_rapport("\nNTFS at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(dump_ind!=0)
	dump(stdscr,buffer,0x200);
      partition->upart_type=UP_NTFS;
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int ntfs_get_attr(const char *attr, const int my_type, t_diskext partition, const char *end)
{
	/* location of first attribute */
	const char *it= attr + NTFS_GETU16(attr + 0x14);
	int type;
	unsigned int len;
	/* Only check for magic DWORD here, fixup should have happened before */
	if(memcmp(attr,"FILE",4)) return 2;
//	aff_buffer(BUFFER_ADD,"FILE\n");
//	aff_buffer(BUFFER_ADD,"seq nbr %lu ",NTFS_GETU16(attr+0x10));
//	aff_buffer(BUFFER_ADD,"main MFT record %lu ",NTFS_GETU64(attr+0x20));
	do
	{
	  unsigned int offset=NTFS_GETU16(it+0x14);
	  const char *data=it+offset;
	  if(data>=end)
	    return 0;
	  type=NTFS_GETU32(it);
	  len=NTFS_GETU16(it+4);
	  if((len<0x18) || (offset>len) || (data+len >= end))
	    return 0;
	  /* We found the attribute type. Is the name correct, too? */
#ifdef NTFS_DEBUG
	  ecrit_rapport("pos=%p type %x data %p\n",it-attr,type,data-attr);
	  dump(stdscr,data,len);
	  aff_buffer(BUFFER_ADD,"type %02X\n",type);
#endif
	  if(type==my_type)
	  {
	    switch(type)
	    {
	      case 0x30:
	      {
		const char *filename_data=data;
		unsigned int namelen=NTFS_GETU8(filename_data+0x40);
		const char *name_it;
		if(filename_data+0x42+namelen>=end)
		  return 0;
//		aff_buffer(BUFFER_ADD,"MFT record nbr %lu ",NTFS_GETU64(filename_data));
		for(name_it=filename_data+0x42;namelen>0; name_it+=2,namelen--)
		  aff_buffer(BUFFER_ADD,"%c",*name_it);
		aff_buffer(BUFFER_ADD,"\n");
	      }
	      break;
	      case 0x60:
		{
		  const char *volumename_data=data;
		  unsigned int namelen=len-offset;
		  const char *name_it;
		  char *dest=partition->name;
		  if(namelen>sizeof(partition->name)-1)
		    namelen=sizeof(partition->name)-1;
		  if(namelen>NTFS_GETU16(it+0x10))
		    namelen=NTFS_GETU16(it+0x10);
		  namelen=namelen/2;
		  for(name_it=volumename_data;(namelen>0) && (*name_it!='\0') && (name_it[1]=='\0'); name_it+=2,namelen--)
		    *dest++=*name_it;
		  *dest++='\0'; // 27 january 2003: Correct a bug found by Andreas du Plessis-Denz
		}
		return 1;
	    }
	  }
	  it+=len;
	}while(type!=-1); /* attribute list end with type -1 */
	return 0;
}

int dir_ntfs(WINDOW *window,t_param_disk disk_car,t_diskext partition)
{
  unsigned char buffer[0x200];
  int debug=0;
  int dump_ind=0;
  aff_buffer(BUFFER_RESET,"Q");
  wmove(window,5,0);
  aff_part(window,1,disk_car,partition);
  aff_part_rapport(1,disk_car,partition);
  ecrit_rapport("\n");
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba+partition->boot_sector)!=0)
  { return 1; }
  if(test_NTFS(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind)!=0)
  {
    aff_buffer(BUFFER_ADD,"Invalid NTFS data\n");
  }
  else
    ntfs_read_MFT(disk_car,partition,(struct ntfs_boot_sector*)&buffer,0x30);
  aff_buffer(BUFFER_DISPLAY,"Q",window);
  return 0;
}

int ntfs_read_MFT(t_param_disk disk_car, t_diskext partition, const struct ntfs_boot_sector*ntfs_header,const int my_type)
{
  unsigned char *buffer;
  unsigned char *attr;
  dword mft_pos;
  unsigned int mft_recordsize;
  mft_pos=partition->lba+ntfs_header->reserved+ntfs_header->mft_cluster*ntfs_header->cluster_size;
  if(ntfs_header->clusters_record>0)
    mft_recordsize=ntfs_header->cluster_size*ntfs_header->clusters_record;
  else
    mft_recordsize=1<<(-ntfs_header->clusters_record);
  buffer=(unsigned char *)MALLOC(mft_recordsize*0x200);
#ifdef NTFS_DEBUG
  ecrit_rapport("NTFS cluster size = %u\n",ntfs_header->cluster_size);
  ecrit_rapport("NTFS MFT_record_size = %u\n",mft_recordsize);
  ecrit_rapport("NTFS MFT cluster = %lu\n",ntfs_header->mft_cluster);
#endif
  if(hd_read_LBA(disk_car,mft_recordsize, buffer, mft_pos)!=0)
  {
    ecrit_rapport(m_ROOT_CLUSTER_RERR);
    FREE(buffer);
    return 1;
  }
  attr=buffer;
  while(attr+0x30<=(buffer+mft_recordsize*0x200))
  {
    if((ntfs_get_attr(attr,my_type,partition,buffer+mft_recordsize*0x200)==2)||
    (NTFS_GETU32(attr + 0x1C)<0x30))
    {
      FREE(buffer);
      return 1;
    }
    attr+= NTFS_GETU32(attr + 0x1C);
  }
  FREE(buffer);
  return 0;
}

int is_ntfs(const int part_type)
{
  switch(part_type)
  {
    case P_NTFS:
    case P_NTFSH:
      return 1;
  }
  return 0;
}

#ifdef NEW_CODE
int create_ntfs_boot_sector(t_param_disk disk_car,const t_diskext partition, const int part_type, const int interface, const __u8 cluster_size, const __u64 mft_cluster, const __u64 mftbak_cluster, const __s32 clusters_frs, const __s32 clusters_record)
{
  unsigned char orgboot[0x200];
  unsigned char newboot[0x200];
  struct ntfs_boot_sector *org_ntfs_header=(struct ntfs_boot_sector *)&orgboot;
  struct ntfs_boot_sector *ntfs_header=(struct ntfs_boot_sector *)&newboot;
  int error=0;
  if(hd_read_LBA(disk_car,1, &orgboot, partition->lba)!=0)
  {
    ecrit_rapport(m_CHKFAT_RERR); return 1;
  }
  memcpy(&newboot,&orgboot,0x200);
  memcpy(ntfs_header->system_id,"NTFS",4);
  ntfs_header->sector_size[0]=0x200 & 0xFF;
  ntfs_header->sector_size[1]=0x200>>8;
  ntfs_header->cluster_size=cluster_size;
  ntfs_header->reserved=0;
  ntfs_header->fats=0;
  ntfs_header->dir_entries[0]=0;
  ntfs_header->dir_entries[1]=0;
  ntfs_header->sectors[0]=0;
  ntfs_header->sectors[1]=0;
  ntfs_header->media=0xF8;
  ntfs_header->fat_length=0;
  ntfs_header->secs_track=disk_car->CHS.sector;
  ntfs_header->heads=disk_car->CHS.head+1;
  if(partition->status==STATUS_LOG)
    ntfs_header->hidden=63;
  else
    ntfs_header->hidden=partition->lba;
  ntfs_header->total_sect=0;
  ntfs_header->sectors_nbr=partition->part_size-1;
  ntfs_header->mft_cluster=mft_cluster;
  ntfs_header->mftbak_cluster=mftbak_cluster;
  ntfs_header->clusters_frs=clusters_frs;
  ntfs_header->clusters_record=clusters_record;
  ntfs_header->marker=0xAA55;
  if(memcmp(newboot,orgboot,0x200))
  {
    ecrit_rapport("             New / Current boot sector");
    dump_2ntfs_rapport(ntfs_header,org_ntfs_header);
    ecrit_rapport("Extrapolated boot sector and current boot sector are different.\n");
  }
  else
  {
    ecrit_rapport("Extrapolated boot sector and current boot sector are identical.\n");
  }
  if(error)
    ecrit_rapport("Bad extrapolation.\n");
  /* */
  if(interface)
  {
    struct MenuItem menuSaveBoot[]=
    {
      { 'Q',"Quit","Quit this section"},
      { 'W', "Write","Write boot"},
      { 0, NULL, NULL }
    };
    const char *options="Q";
    int do_write=0;
    aff_copy(stdscr);
    move(4,0);
    aff_disk(stdscr,disk_car);
    mvaddstr(6,0,m_PART_HEADER2);
    move(7,0);
    aff_part(stdscr,2,disk_car,partition);
    move(8,0);
    if(memcmp(newboot,orgboot,0x200))	/* Only compare the first sector */
    {
      options="WQ";
 //     dump_2ntfs_info(ntfs_header, org_ntfs_header, part_type);
      doprintf("Extrapolated boot sector and current boot sector are different.\n");
      if(error)
	ecrit_rapport("Warning: Extrapolated boot sector have incorrect values.\n");
    }
    else
    {
      dump_ntfs_info(ntfs_header);
      doprintf("Extrapolated boot sector and current boot sector are identical.\n");
    }
    switch(toupper(menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuSaveBoot,8,options,MENU_HORIZ | MENU_ACCEPT_OTHERS, 0)))
    {
      case 'W':
	if(strchr(options,'W'))
	  do_write=1;
	break;
    }
    if(do_write && ask_confirmation("Confirm ? (Y/N)"))
    {
      ecrit_rapport("Write new boot!\n");
      /* Write boot sector and backup boot sector */
      hd_write_LBA(disk_car,1, &newboot, partition->lba);
      hd_write_LBA(disk_car,1, &newboot, partition->lba+partition->part_size-1);
    }
    else
      ecrit_rapport("Don't write new boot!\n");
  }
  return 0;
}

int dump_2ntfs_info(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2)
{
  doprintf("cluster_size    %u %u\n",nh1->cluster_size,nh2->cluster_size);
  doprintf("mft_cluster     %lu %lu\n",nh1->mft_cluster,nh2->mft_cluster);
  doprintf("mftbak_cluster  %lu %lu\n",nh1->mftbak_cluster,nh2->mftbak_cluster);
  doprintf("clusters_frs    %d %d\n",nh1->clusters_frs,nh2->clusters_frs);
  doprintf("clusters_record %d %d\n",nh1->clusters_record,nh2->clusters_record);
}

int dump_2ntfs_rapport(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2)
{
  ecrit_rapport("cluster_size    %u %u\n",nh1->cluster_size,nh2->cluster_size);
  ecrit_rapport("mft_cluster     %lu %lu\n",nh1->mft_cluster,nh2->mft_cluster);
  ecrit_rapport("mftbak_cluster  %lu %lu\n",nh1->mftbak_cluster,nh2->mftbak_cluster);
  ecrit_rapport("clusters_frs    %d %d\n",nh1->clusters_frs,nh2->clusters_frs);
  ecrit_rapport("clusters_record %d %d\n",nh1->clusters_record,nh2->clusters_record);
}

int dump_ntfs_info(const struct ntfs_boot_sector *ntfs_header)
{
  doprintf("cluster_size    %u\n",ntfs_header->cluster_size);
  doprintf("mft_cluster     %lu\n",ntfs_header->mft_cluster);
  doprintf("mftbak_cluster  %lu\n",ntfs_header->mftbak_cluster);
  doprintf("clusters_frs    %d\n",ntfs_header->clusters_frs);
  doprintf("clusters_record %d\n",ntfs_header->clusters_record);
}
#endif
